# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import base64
import json
import os
import sqlalchemy
import tempfile
import time
import zipfile
import zlib

from collections import defaultdict
from datetime import datetime, timedelta
from hashlib import sha256
from tempfile import TemporaryDirectory
from typing import Any, Dict, List, Optional, Set, Tuple, Union, cast

import codechecker_api_shared
from codechecker_api.codeCheckerDBAccess_v6 import ttypes

from codechecker_common import skiplist_handler
from codechecker_common.logger import get_logger
from codechecker_common.review_status_handler import ReviewStatusHandler, \
    SourceReviewStatus
from codechecker_common.util import load_json, path_for_fake_root

from codechecker_report_converter.util import trim_path_prefixes
from codechecker_report_converter.report import \
    FakeChecker, Report, UnknownChecker, report_file
from codechecker_report_converter.report.hash import get_report_path_hash

from ..database import db_cleanup
from ..database.config_db_model import Product
from ..database.database import DBSession
from ..database.run_db_model import \
    AnalysisInfo, AnalysisInfoChecker, AnalyzerStatistic, \
    BugPathEvent, BugReportPoint, \
    Checker, \
    ExtendedReportData, \
    File, FileContent, \
    Report as DBReport, ReportAnnotations, ReviewStatus as ReviewStatusRule, \
    Run, RunLock, RunHistory
from ..metadata import checker_is_unavailable, MetadataInfoParser

from .report_server import ThriftRequestHandler
from .thrift_enum_helper import report_extended_data_type_str


LOG = get_logger('server')
STORE_TIME_LOG = get_logger('store_time')


class LogTask:
    def __init__(self, run_name: str, message: str):
        self.__run_name = run_name
        self.__msg = message
        self.__start_time = time.time()

    def __enter__(self, *args):
        LOG.info("[%s] %s...", self.__run_name, self.__msg)

    def __exit__(self, *args):
        LOG.info("[%s] %s. Done. (Duration: %s sec)", self.__run_name,
                 self.__msg, round(time.time() - self.__start_time, 2))


class RunLocking:
    def __init__(self, session: DBSession, run_name: str):
        self.__session = session
        self.__run_name = run_name

    def __enter__(self, *args):
        # Load the lock record for "FOR UPDATE" so that the transaction that
        # handles the run's store operations has a lock on the database row
        # itself.
        self.__run_lock = self.__session.query(RunLock) \
            .filter(RunLock.name == self.__run_name) \
            .with_for_update(nowait=True) \
            .one()

        # Do *NOT* remove this seemingly dummy print. The query functions on
        # result in a **proxy** object, essentially a weak pointer, without(!)
        # the execution of any underlying SQL statements. Which means that
        # without the evaluation of the member of this query (with the
        # . operator) there would be no actual DBMS-level lock in place.
        # (Not having debug logs enabled is not a problem: the evaluation of
        # the arguments still take place even if the logging configuration
        # prevents the actual printing of the log line.)
        LOG.debug("Acquired exclusive lock for run '%s' that was originally "
                  "locked at '%s'.",
                  self.__run_name, self.__run_lock.locked_at)
        return self

    def __exit__(self, *args):
        self.__run_lock = None
        self.__session = None


def unzip(b64zip: str, output_dir: str) -> int:
    """
    This function unzips the base64 encoded zip file. This zip is extracted
    to a temporary directory and the ZIP is then deleted. The function returns
    the size of the extracted decompressed zip file.
    """
    if len(b64zip) == 0:
        return 0

    with tempfile.NamedTemporaryFile(suffix='.zip') as zip_file:
        LOG.debug("Unzipping mass storage ZIP '%s' to '%s'...",
                  zip_file.name, output_dir)

        zip_file.write(zlib.decompress(base64.b64decode(b64zip)))
        with zipfile.ZipFile(zip_file, 'r', allowZip64=True) as zipf:
            try:
                zipf.extractall(output_dir)
                return os.stat(zip_file.name).st_size
            except Exception:
                LOG.error("Failed to extract received ZIP.")
                import traceback
                traceback.print_exc()
                raise
    return 0


def get_file_content(file_path: str) -> bytes:
    """Return the file content for the given filepath. """
    with open(file_path, 'rb') as f:
        return f.read()


def add_file_record(
    session: DBSession,
    file_path: str,
    content_hash: str
) -> Optional[int]:
    """
    Add the necessary file record pointing to an already existing content.
    Returns the added file record id or None, if the content_hash is not
    found.

    This function must not be called between add_checker_run() and
    finish_checker_run() functions when SQLite database is used!
    add_checker_run() function opens a transaction which is closed by
    finish_checker_run() and since SQLite doesn't support parallel
    transactions, this API call will wait until the other transactions
    finish. In the meantime the run adding transaction times out.

    This function doesn't insert blame info in the File objects because those
    are added by __add_blame_info(). In previous CodeChecker versions git info
    was not stored at all, so upgrading to a new CodeChecker results the
    storage of git blame info even for files that are already stored. For this
    reason we can't avoid an update on File and FileContent tables, but we can
    avoid double reading of blame info json files.
    """
    file_record = session.query(File) \
        .filter(File.content_hash == content_hash,
                File.filepath == file_path) \
        .one_or_none()

    if file_record:
        return file_record.id

    try:
        # Parallel storage of runs containing common file paths results a
        # "duplicate key violation" error. This is handled by CodeChecker, so
        # practically it causes no problem. The INSERT command of the second
        # transaction will be thrown away. However, some DB systems are
        # supporting "ON CONFLICT DO NOTHING" clause in INSERT statement which
        # solves the same issue gracefully.
        # TODO: "ON CONFLICT DO NOTHING" feature is available for SQLite engine
        # too in SQLAlchemy 1.4.
        if session.bind.dialect.name == 'postgresql':
            insert_stmt = sqlalchemy.dialects.postgresql.insert(File).values(
                filepath=file_path,
                filename=os.path.basename(file_path),
                content_hash=content_hash).on_conflict_do_nothing(
                    index_elements=['filepath', 'content_hash'])
            file_id = session.execute(insert_stmt).inserted_primary_key[0]
            session.commit()
            return file_id
        else:
            file_record = File(file_path, content_hash, None, None)
            session.add(file_record)
        session.commit()
    except sqlalchemy.exc.IntegrityError as ex:
        LOG.error(ex)
        # Other transaction might have added the same file in the
        # meantime.
        session.rollback()
        file_record = session.query(File) \
            .filter(File.content_hash == content_hash,
                    File.filepath == file_path).one_or_none()

    return file_record.id if file_record else None


def get_blame_file_data(
    blame_file: str
) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Get blame information from the given file.

    It will return a tuple of 'blame information', 'remote url' and
    'tracking branch'.
    """
    blame_info = None
    remote_url = None
    tracking_branch = None

    if os.path.isfile(blame_file):
        data = load_json(blame_file)
        if data:
            remote_url = data.get("remote_url")
            tracking_branch = data.get("tracking_branch")

            blame_info = data

            # Remove fields which are not needed anymore from the blame info.
            del blame_info["remote_url"]
            del blame_info["tracking_branch"]

    return blame_info, remote_url, tracking_branch


def checker_name_for_report(report: Report) -> Tuple[str, str]:
    return (report.analyzer_name or UnknownChecker[0],
            report.checker_name or UnknownChecker[1])


class MassStoreRun:
    def __init__(
        self,
        report_server: ThriftRequestHandler,
        name: str,
        tag: Optional[str],
        version: Optional[str],
        b64zip: str,
        force: bool,
        trim_path_prefixes: Optional[List[str]],
        description: Optional[str]
    ):
        """ Initialize object. """
        self.__report_server = report_server

        self.__name = name
        self.__tag = tag
        self.__version = version
        self.__b64zip = b64zip
        self.__force = force
        self.__trim_path_prefixes = trim_path_prefixes
        self.__description = description

        self.__mips: Dict[str, MetadataInfoParser] = dict()
        self.__analysis_info: Dict[str, AnalysisInfo] = dict()
        self.__checker_row_cache: Dict[Tuple[str, str], Checker] = dict()
        self.__duration: int = 0
        self.__report_count: int = 0
        self.__report_limit: int = 0
        self.__wrong_src_code_comments: List[str] = list()
        self.__already_added_report_hashes: Set[str] = set()
        self.__new_report_hashes: Dict[str, Tuple] = dict()
        self.__all_report_checkers: Set[str] = set()
        self.__added_reports: List[Tuple[DBReport, Report]] = list()
        self.__reports_with_fake_checkers: Dict[
            # Either a DBReport *without* an ID, or the ID of a committed
            # DBReport.
            str, Tuple[Report, Union[DBReport, int]]] = dict()

        self.__get_report_limit_for_product()

    @property
    def __manager(self):
        return self.__report_server._manager

    @property
    def __Session(self):
        return self.__report_server._Session

    @property
    def __config_database(self):
        return self.__report_server._config_database

    @property
    def __product(self):
        return self.__report_server._product

    @property
    def __context(self):
        return self.__report_server._context

    @property
    def user_name(self):
        return self.__report_server._get_username()

    def __check_run_limit(self):
        """
        Checks the maximum allowed of uploadable runs for the current product.
        """
        max_run_count = self.__manager.get_max_run_count()

        with DBSession(self.__config_database) as session:
            product = session.query(Product).get(self.__product.id)
            if product.run_limit:
                max_run_count = product.run_limit

        # Session that handles constraints on the run.
        with DBSession(self.__Session) as session:
            if not max_run_count:
                return

            LOG.debug("Check the maximum number of allowed runs which is %d",
                      max_run_count)

            run = session.query(Run) \
                .filter(Run.name == self.__name) \
                .one_or_none()

            # If max_run_count is not set in the config file, it will allow
            # the user to upload unlimited runs.

            run_count = session.query(Run.id).count()

            # If we are not updating a run or the run count is reached the
            # limit it will throw an exception.
            if not run and run_count >= max_run_count:
                remove_run_count = run_count - max_run_count + 1
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                    f"You reached the maximum number of allowed runs "
                    f"({run_count}/{max_run_count})! Please remove at least "
                    f"{remove_run_count} run(s) before you try it again.")

    def __store_run_lock(self, session: DBSession):
        """
        Store a RunLock record for the given run name into the database.
        """
        try:
            # If the run can be stored, we need to lock it first. If there is
            # already a lock in the database for the given run name which is
            # expired and multiple processes are trying to get this entry from
            # the database for update we may get the following exception:
            # could not obtain lock on row in relation "run_locks"
            # This is the reason why we have to wrap this query to a try/except
            # block.
            run_lock = session.query(RunLock) \
                .filter(RunLock.name == self.__name) \
                .with_for_update(nowait=True).one_or_none()
        except (sqlalchemy.exc.OperationalError,
                sqlalchemy.exc.ProgrammingError) as ex:
            LOG.error("Failed to get run lock for '%s': %s", self.__name, ex)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                "Someone is already storing to the same run. Please wait "
                "while the other storage is finished and try it again.")

        if not run_lock:
            # If there is no lock record for the given run name, the run
            # is not locked -- create a new lock.
            run_lock = RunLock(self.__name, self.user_name)
            session.add(run_lock)
        elif run_lock.has_expired(
                db_cleanup.RUN_LOCK_TIMEOUT_IN_DATABASE):
            # There can be a lock in the database, which has already
            # expired. In this case, we assume that the previous operation
            # has failed, and thus, we can re-use the already present lock.
            run_lock.touch()
            run_lock.username = self.user_name
        else:
            # In case the lock exists and it has not expired, we must
            # consider the run a locked one.
            when = run_lock.when_expires(
                db_cleanup.RUN_LOCK_TIMEOUT_IN_DATABASE)

            username = run_lock.username if run_lock.username is not None \
                else "another user"

            LOG.info("Refusing to store into run '%s' as it is locked by "
                     "%s. Lock will expire at '%s'.", self.__name, username,
                     when)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                "The run named '{0}' is being stored into by {1}. If the "
                "other store operation has failed, this lock will expire "
                "at '{2}'.".format(self.__name, username, when))

        # At any rate, if the lock has been created or updated, commit it
        # into the database.
        try:
            session.commit()
        except (sqlalchemy.exc.IntegrityError,
                sqlalchemy.orm.exc.StaleDataError):
            # The commit of this lock can fail.
            #
            # In case two store ops attempt to lock the same run name at the
            # same time, committing the lock in the transaction that commits
            # later will result in an IntegrityError due to the primary key
            # constraint.
            #
            # In case two store ops attempt to lock the same run name with
            # reuse and one of the operation hangs long enough before COMMIT
            # so that the other operation commits and thus removes the lock
            # record, StaleDataError is raised. In this case, also consider
            # the run locked, as the data changed while the transaction was
            # waiting, as another run wholly completed.

            LOG.info("Run '%s' got locked while current transaction "
                     "tried to acquire a lock. Considering run as locked.",
                     self.__name)
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                "The run named '{0}' is being stored into by another "
                "user.".format(self.__name))

    def __free_run_lock(self, session: DBSession):
        """ Remove the lock from the database for the given run name. """
        # Using with_for_update() here so the database (in case it supports
        # this operation) locks the lock record's row from any other access.
        run_lock = session.query(RunLock) \
            .filter(RunLock.name == self.__name) \
            .with_for_update(nowait=True).one()
        session.delete(run_lock)
        session.commit()

    def __store_source_files(
        self,
        source_root: str,
        filename_to_hash: Dict[str, str]
    ) -> Dict[str, int]:
        """ Storing file contents from plist. """

        file_path_to_id = {}

        for file_name, file_hash in filename_to_hash.items():
            source_file_path = path_for_fake_root(file_name, source_root)
            LOG.debug("Storing source file: %s", source_file_path)
            trimmed_file_path = trim_path_prefixes(
                file_name, self.__trim_path_prefixes)

            if not os.path.isfile(source_file_path):
                # The file was not in the ZIP file, because we already
                # have the content. Let's check if we already have a file
                # record in the database or we need to add one.

                LOG.debug('%s not found or already stored.', trimmed_file_path)
                with DBSession(self.__Session) as session:
                    fid = add_file_record(
                        session, trimmed_file_path, file_hash)

                if fid:
                    file_path_to_id[trimmed_file_path] = fid
                    LOG.debug("%d fileid found", fid)
                else:
                    LOG.error("File ID for %s is not found in the DB with "
                              "content hash %s. Missing from ZIP?",
                              source_file_path, file_hash)
                continue

            with DBSession(self.__Session) as session:
                self.__add_file_content(session, source_file_path, file_hash)

                file_path_to_id[trimmed_file_path] = add_file_record(
                    session, trimmed_file_path, file_hash)

        return file_path_to_id

    def __add_blame_info(
        self,
        blame_root: str,
        filename_to_hash: Dict[str, str]
    ):
        """
        This function updates blame info in File and FileContent tables if
        these were NULL. This function exists only because in earlier
        CodeChecker versions blame info was not stored in these tables and
        in a subsequent storage we can't update the tables for an unchanged
        file since __store_source_files() updates only the ones that are in the
        .zip file. This function stores blame info even if the corresponding
        source file is not in the .zip file.
        """
        with DBSession(self.__Session) as session:
            for subdir, _, files in os.walk(blame_root):
                for f in files:
                    blame_file = os.path.join(subdir, f)
                    file_path = blame_file[len(blame_root.rstrip("/")):]
                    blame_info, remote_url, tracking_branch = \
                        get_blame_file_data(blame_file)

                    compressed_blame_info = None
                    if blame_info:
                        compressed_blame_info = zlib.compress(
                            json.dumps(blame_info).encode('utf-8'),
                            zlib.Z_BEST_COMPRESSION)

                    session \
                        .query(FileContent) \
                        .filter(FileContent.blame_info.is_(None)) \
                        .filter(FileContent.content_hash ==
                                filename_to_hash.get(file_path)) \
                        .update({"blame_info": compressed_blame_info})

                    session \
                        .query(File) \
                        .filter(File.remote_url.is_(None)) \
                        .filter(File.filepath == file_path) \
                        .update({
                            "remote_url": remote_url,
                            "tracking_branch": tracking_branch})

            session.commit()

    def __add_file_content(
        self,
        session: DBSession,
        source_file_name: str,
        content_hash: Optional[str]
    ):
        """
        Add the necessary file contents. If content_hash in None then this
        function calculates the content hash. Or if it's available at the
        caller and it's provided then it will not be calculated again.

        This function must not be called between add_checker_run() and
        finish_checker_run() functions when SQLite database is used!
        add_checker_run() function opens a transaction which is closed by
        finish_checker_run() and since SQLite doesn't support parallel
        transactions, this API call will wait until the other transactions
        finish. In the meantime the run adding transaction times out.

        This function doesn't insert blame info in the FileContent objects
        because those are added by __add_blame_info(). In previous CodeChecker
        versions git info was not stored at all, so upgrading to a new
        CodeChecker results the storage of git blame info even for files that
        are already stored. For this reason we can't avoid an update on File
        and FileContent tables, but we can avoid double reading of blame info
        json files.
        """
        source_file_content = None
        if not content_hash:
            source_file_content = get_file_content(source_file_name)

            hasher = sha256()
            hasher.update(source_file_content)
            content_hash = hasher.hexdigest()

        file_content = session.query(FileContent).get(content_hash)
        if not file_content:
            if not source_file_content:
                source_file_content = get_file_content(source_file_name)
            try:
                compressed_content = zlib.compress(
                    source_file_content, zlib.Z_BEST_COMPRESSION)

                if session.bind.dialect.name == 'postgresql':
                    insert_stmt = sqlalchemy.dialects.postgresql \
                        .insert(FileContent).values(
                            content_hash=content_hash,
                            content=compressed_content,
                            blame_info=None).on_conflict_do_nothing(
                                index_elements=['content_hash'])

                    session.execute(insert_stmt)
                else:
                    fc = FileContent(content_hash, compressed_content, None)
                    session.add(fc)

                session.commit()
            except sqlalchemy.exc.IntegrityError:
                # Other transaction moght have added the same content in
                # the meantime.
                session.rollback()

    def __store_checker_identifiers(self, checkers: Set[Tuple[str, str]]):
        """
        Stores the identifiers "(analyzer, checker_name)" in the database into
        a look-up table where each unique checker is given a unique numeric
        identifier.

        Due to the use of an M-to-N connection table
        (see `AnalysisInfoChecker`) one side of the joins must have their IDs
        eagerly populated, otherwise the Python bindings will fail.
        However, eager population will result in exceptions that a flush was
        created before the transaction was complete.

        Moreover, this is performed separately from the storing of the details
        of a run to reduce contention if two parallel stores, especially across
        server instances (in a distributed/load-balanced environment) want to
        store the same identifier(s).
        """
        max_tries, tries, wait_time = 3, 0, timedelta(seconds=30)
        # The "fake" checker is a temporary row that is needed intermittently
        # during report storage because there might be reports that point to
        # checkers which are not found in a preemptively parsed
        # 'metadata.json', or, in the worst case, there might simply not be
        # a 'metadata.json' at all in the to-be-stored structure.
        all_checkers = {FakeChecker, UnknownChecker} | checkers
        while tries < max_tries:
            tries += 1
            try:
                LOG.debug("[%s] Begin attempt %d...", self.__name, tries)
                with DBSession(self.__Session) as session:
                    known_checkers = {(r.analyzer_name, r.checker_name)
                                      for r in session
                                      .query(Checker.analyzer_name,
                                             Checker.checker_name)
                                      .all()}
                    for analyzer, checker in \
                            sorted(all_checkers - known_checkers):
                        s = self.__context.checker_labels.severity(checker)
                        s = ttypes.Severity._NAMES_TO_VALUES[s]
                        session.add(Checker(analyzer, checker, s))
                        LOG.debug("Acquiring ID for checker '%s/%s' "
                                  "for the first time.", analyzer, checker)

                    session.commit()
                    return
            except (sqlalchemy.exc.OperationalError,
                    sqlalchemy.exc.ProgrammingError) as ex:
                LOG.error("Storing checkers of run '%s' failed: %s.\n"
                          "Waiting %d before trying again...",
                          self.__name, ex, wait_time)
                time.sleep(wait_time.total_seconds())
                wait_time *= 2
            except Exception as ex:
                LOG.error("Failed to store checkers due to some other error: "
                          "%s", ex)
                import traceback
                traceback.print_exc()
                raise

        raise codechecker_api_shared.ttypes.RequestFailed(
            codechecker_api_shared.ttypes.ErrorCode.DATABASE,
            "Storing the names of the checkers in the run failed due to "
            "excessive contention!")

    def __store_analysis_statistics(
        self,
        session: DBSession,
        run_history_id: int
    ):
        """
        Store analysis statistics for the given run history.

        It will unique the statistics for each analyzer type based on the
        metadata information.
        """
        stats = defaultdict(lambda: {
            "versions": set(),
            "failed_sources": set(),
            "successful_sources": set(),
            "successful": 0
        })

        for mip in self.__mips.values():
            self.__duration += int(sum(mip.check_durations))

            for analyzer_type, res in mip.analyzer_statistics.items():
                if "version" in res:
                    stats[analyzer_type]["versions"].add(res["version"])

                if "failed_sources" in res:
                    if self.__version == '6.9.0':
                        stats[analyzer_type]["failed_sources"].add(
                            'Unavailable in CodeChecker 6.9.0!')
                    else:
                        stats[analyzer_type]["failed_sources"].update(
                            res["failed_sources"])

                if "successful_sources" in res:
                    stats[analyzer_type]["successful_sources"].update(
                        res["successful_sources"])

                if "successful" in res:
                    stats[analyzer_type]["successful"] += res["successful"]

        for analyzer_type, stat in stats.items():
            analyzer_version = None
            if stat["versions"]:
                analyzer_version = zlib.compress(
                    "; ".join(stat["versions"]).encode('utf-8'),
                    zlib.Z_BEST_COMPRESSION)

            failed = 0
            compressed_files = None
            if stat["failed_sources"]:
                compressed_files = zlib.compress(
                    '\n'.join(stat["failed_sources"]).encode('utf-8'),
                    zlib.Z_BEST_COMPRESSION)

                failed = len(stat["failed_sources"])

            successful = len(stat["successful_sources"]) \
                if stat["successful_sources"] else stat["successful"]

            analyzer_statistics = AnalyzerStatistic(
                run_history_id, analyzer_type, analyzer_version,
                successful, failed, compressed_files)

            session.add(analyzer_statistics)

    def __store_analysis_info(
        self,
        session: DBSession,
        run_history: RunHistory
    ):
        """ Store analysis info for the given run history. """
        for src_dir_path, mip in self.__mips.items():
            for analyzer_command in mip.check_commands:
                cmd = zlib.compress(
                    analyzer_command.encode("utf-8"),
                    zlib.Z_BEST_COMPRESSION)

                analysis_info_rows = session \
                    .query(AnalysisInfo) \
                    .filter(AnalysisInfo.analyzer_command == cmd) \
                    .all()

                if analysis_info_rows:
                    # It is possible when multiple runs are stored
                    # simultaneously to the server with the same analysis
                    # command that multiple entries are stored into the
                    # database. In this case we will select the first one.
                    analysis_info = analysis_info_rows[0]
                else:
                    analysis_info = AnalysisInfo(analyzer_command=cmd)

                    # Obtain the ID eagerly to be able to use the M-to-N table.
                    session.add(analysis_info)
                    session.flush()
                    session.refresh(analysis_info, ["id"])

                    for analyzer in mip.analyzers:
                        q = session \
                            .query(Checker) \
                            .filter(Checker.analyzer_name == analyzer)
                        db_checkers = {r.checker_name: r for r in q.all()}

                        connection_rows = [AnalysisInfoChecker(
                            analysis_info, db_checkers[chk], is_enabled)
                            for chk, is_enabled
                            in mip.checkers.get(analyzer, dict()).items()]
                        for r in connection_rows:
                            session.add(r)

                run_history.analysis_info.append(analysis_info)
                self.__analysis_info[src_dir_path] = analysis_info

    def __add_or_update_run(
        self,
        session: DBSession,
        run_history_time: datetime
    ) -> Tuple[int, bool]:
        """
        Store run related data to the database.
        By default updates the results if name already exists.
        Using the force flag removes existing analysis results for a run.
        """
        try:
            LOG.debug("Adding run '%s'...", self.__name)

            run = session.query(Run) \
                .filter(Run.name == self.__name) \
                .one_or_none()

            update_run = True
            if run and self.__force:
                # Clean already collected results.
                if not run.can_delete:
                    # Deletion is already in progress.
                    msg = f"Can't delete {run.id}"
                    LOG.debug(msg)
                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                        msg)

                LOG.info('Removing previous analysis results...')
                session.delete(run)
                # Not flushing after delete leads to a constraint violation
                # error later, when adding run entity with the same name as
                # the old one.
                session.flush()

                checker_run = Run(self.__name, self.__version)
                session.add(checker_run)
                session.flush()
                run_id = checker_run.id

            elif run:
                # There is already a run, update the results.
                run.date = datetime.now()
                run.duration = -1
                session.flush()
                run_id = run.id
            else:
                # There is no run create new.
                checker_run = Run(self.__name, self.__version)
                session.add(checker_run)
                session.flush()
                run_id = checker_run.id
                update_run = False

            # Add run to the history.
            LOG.debug("Adding run history.")

            if self.__tag is not None:
                run_history = session.query(RunHistory) \
                    .filter(RunHistory.run_id == run_id,
                            RunHistory.version_tag == self.__tag) \
                    .one_or_none()

                if run_history:
                    run_history.version_tag = None
                    session.add(run_history)

            cc_versions = set()
            for mip in self.__mips.values():
                if mip.cc_version:
                    cc_versions.add(mip.cc_version)

            cc_version = '; '.join(cc_versions) if cc_versions else None
            run_history = RunHistory(
                run_id, self.__tag, self.user_name, run_history_time,
                cc_version, self.__description)

            session.add(run_history)
            session.flush()

            LOG.debug("Adding run done.")

            self.__store_analysis_statistics(session, run_history.id)
            self.__store_analysis_info(session, run_history)

            session.flush()
            LOG.debug("Storing analysis statistics done.")

            return run_id, update_run
        except Exception as ex:
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                str(ex))

    def __get_checker(self,
                      session: DBSession,
                      analyzer_name: str,
                      checker_name: str) -> Optional[Checker]:
        try:
            return self.__checker_row_cache[(analyzer_name, checker_name)]
        except KeyError:
            maybe_orm: Optional[Checker] = session.query(Checker) \
                .filter(sqlalchemy.and_(
                    Checker.analyzer_name == analyzer_name,
                    Checker.checker_name == checker_name)) \
                .one_or_none()
            if maybe_orm:
                self.__checker_row_cache[(analyzer_name, checker_name)] = \
                    cast(Checker, maybe_orm)
            return maybe_orm

    def __checker_for_report(self,
                             session: DBSession,
                             report: Report) -> Optional[Checker]:
        analyzer_name, checker_name = checker_name_for_report(report)
        return self.__get_checker(session, analyzer_name, checker_name)

    def __add_report(
        self,
        session: DBSession,
        run_id: int,
        report: Report,
        report_path_hash: str,
        file_path_to_id: Dict[str, int],
        review_status: SourceReviewStatus,
        detection_status: str,
        detection_time: datetime,
        run_history_time: datetime,
        analysis_info: Optional[AnalysisInfo],
        fixed_at: Optional[datetime] = None
    ) -> int:
        """ Add report to the database. """
        try:
            checker = self.__checker_for_report(session, report)
            if not checker:
                # It would be too easy to create a 'Checker' instance with the
                # observed data right here, but __add_report() is called in
                # the context of the *BIG* TRANSACTION which has all the
                # reports of the entire store pending. Losing all that
                # information on a potential UNIQUE CONSTRAINT violation due
                # to multiple concurrent massStoreRun()s trying to store the
                # same checker ID which was never seen in a 'metadata.json' is
                # not worth it.
                checker = self.__get_checker(session,
                                             FakeChecker[0], FakeChecker[1])
                if not checker:
                    LOG.fatal("Psuedo-checker '%s/%s' has no "
                              "identity in the database, even though "
                              "__store_checker_identifiers() should have "
                              "always preemptively created it!",
                              FakeChecker[0], FakeChecker[1])
                    raise KeyError(FakeChecker[1])

            db_report = DBReport(
                file_path_to_id[report.file.path], run_id, report.report_hash,
                checker, report.line, report.column,
                len(report.bug_path_events), report.message, detection_status,
                review_status.status, review_status.author,
                review_status.message, run_history_time,
                review_status.in_source, detection_time, fixed_at)
            if analysis_info:
                db_report.analysis_info.append(analysis_info)

            session.add(db_report)
            self.__added_reports.append((db_report, report))
            if db_report.checker.checker_name == FakeChecker[1]:
                self.__reports_with_fake_checkers[report_path_hash] = \
                    (report, db_report)

            return db_report.id

        except Exception as ex:
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                str(ex))

    def __get_faked_checkers(self) \
            -> Set[Tuple[str, str]]:
        """
        Extracts the "real" checker identifiers from the
        __reports_with_fake_checkers that might contain some yet not fully
        handled reports by __add_report(). This function does NOT touch the
        database!
        """
        return set(checker_name_for_report(report)
                   for report, _
                   in self.__reports_with_fake_checkers.values())

    def __load_report_ids_for_reports_with_fake_checkers(self, session):
        """
        Transforms the __reports_with_fake_checkers data structure by loading
        the report.id column in place for the DBReport ORM objects to allow
        __realise_fake_checkers() to execute appropriately.

        This must only be run **once** between the __add_report() sequence and
        the __realise_fake_checkers() operation. The reports **MUST** have
        been committed prior.
        """
        for rph in self.__reports_with_fake_checkers:
            report, db_report = cast(Tuple[Report, DBReport],
                                     self.__reports_with_fake_checkers[rph])
            # Only load the "id" field from the database, not the entire row.
            session.refresh(db_report, ["id"])
            id_: int = db_report.id

            self.__reports_with_fake_checkers[rph] = (report, id_)

    def __realise_fake_checkers(self, session):
        """
        __add_report() might leave some reports that have checker names in
        their data that were not found in the usually full checker list
        available in the 'metadata.json'. This usually happens if the report
        directory that is being stored is not from an 'analyze' or 'check'
        invocation (that would create an appropriate 'metadata.json') but from
        other tools (e.g., report-converter, or Cppcheck's native PLISTs).

        This function assumes that for each report in the
        __reports_with_fake_checkers, their actual checkers' IDs are now added
        added to the database (which was not the case when __add_report() ran),
        so all it does is upgrade the 'checker_id' FOREIGN KEY field to point
        at the real checker.
        """
        try:
            grouped_by_checker: Dict[Tuple[str, str], List[int]] = \
                defaultdict(list)
            for _, (report, db_id) in \
                    self.__reports_with_fake_checkers.items():
                checker: Tuple[str, str] = checker_name_for_report(report)
                grouped_by_checker[checker].append(cast(int, db_id))

            for chk, report_ids in grouped_by_checker.items():
                analyzer_name, checker_name = chk
                chk_obj = cast(Checker, self.__get_checker(session,
                                                           analyzer_name,
                                                           checker_name))
                session.query(DBReport) \
                    .filter(DBReport.id.in_(report_ids)) \
                    .update({"checker_id": chk_obj.id},
                            synchronize_session=False)
        except Exception as ex:
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                str(ex))

    def __add_report_context(self, session, file_path_to_id):
        try:
            for db_report, report in self.__added_reports:
                LOG.debug("Storing bug path positions.")
                for i, p in enumerate(report.bug_path_positions):
                    session.add(BugReportPoint(
                        p.range.start_line, p.range.start_col,
                        p.range.end_line, p.range.end_col,
                        i, file_path_to_id[p.file.path], db_report.id))

                LOG.debug("Storing bug path events.")
                for i, event in enumerate(report.bug_path_events):
                    session.add(BugPathEvent(
                        event.range.start_line, event.range.start_col,
                        event.range.end_line, event.range.end_col,
                        i, event.message, file_path_to_id[event.file.path],
                        db_report.id))

                LOG.debug("Storing notes.")
                for note in report.notes:
                    data_type = report_extended_data_type_str(
                        ttypes.ExtendedReportDataType.NOTE)

                    session.add(ExtendedReportData(
                        note.range.start_line, note.range.start_col,
                        note.range.end_line, note.range.end_col,
                        note.message, file_path_to_id[note.file.path],
                        db_report.id, data_type))

                LOG.debug("Storing macro expansions.")
                for macro in report.macro_expansions:
                    data_type = report_extended_data_type_str(
                        ttypes.ExtendedReportDataType.MACRO)

                    session.add(ExtendedReportData(
                        macro.range.start_line, macro.range.start_col,
                        macro.range.end_line, macro.range.end_col,
                        macro.message, file_path_to_id[macro.file.path],
                        db_report.id, data_type))

                if report.annotations:
                    self.__validate_and_add_report_annotations(
                        session, db_report.id, report.annotations)

            session.flush()

        except Exception as ex:
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.ErrorCode.GENERAL,
                str(ex))

    def __process_report_file(
        self,
        report_file_path: str,
        session: DBSession,
        run_id: int,
        file_path_to_id: Dict[str, int],
        run_history_time: datetime,
        skip_handler: skiplist_handler.SkipListHandler,
        review_status_handler: ReviewStatusHandler,
        hash_map_reports: Dict[str, List[Any]]
    ) -> bool:
        """
        Process and save reports from the given report file to the database.
        """
        reports = report_file.get_reports(report_file_path)

        if not reports:
            return True

        def get_missing_file_ids(report: Report) -> List[str]:
            """ Returns file paths which database file id is missing. """
            missing_ids_for_files = []
            for file_path in report.trimmed_files:
                file_id = file_path_to_id.get(file_path, -1)
                if file_id == -1:
                    missing_ids_for_files.append(file_path)

            return missing_ids_for_files

        root_dir_path = os.path.dirname(report_file_path)
        mip = self.__mips[root_dir_path]
        analysis_info = self.__analysis_info.get(root_dir_path)

        for report in reports:
            self.__report_count += 1
            report.trim_path_prefixes(self.__trim_path_prefixes)

            missing_ids_for_files = get_missing_file_ids(report)
            if missing_ids_for_files:
                LOG.warning("Failed to get database id for file path '%s'! "
                            "Skip adding report: %s:%d:%d [%s]",
                            ' '.join(missing_ids_for_files), report.file.path,
                            report.line, report.column, report.checker_name)
                continue

            self.__all_report_checkers.add(report.checker_name)

            if skip_handler.should_skip(report.file.original_path):
                continue

            report_path_hash = get_report_path_hash(report)
            if report_path_hash in self.__already_added_report_hashes:
                LOG.debug('Not storing report. Already added: %s', report)
                continue

            LOG.debug("Storing report to the database...")

            detection_status = 'new'
            detected_at = run_history_time

            old_report = None
            if report.report_hash in hash_map_reports:
                old_report = hash_map_reports[report.report_hash][0]
                old_status = old_report.detection_status
                detection_status = 'reopened' \
                    if old_status == 'resolved' else 'unresolved'
                detected_at = old_report.detected_at

            report.analyzer_name = mip.checker_to_analyzer.get(
                report.checker_name, report.analyzer_name)

            review_status = SourceReviewStatus()

            try:
                review_status = review_status_handler.get_review_status(report)
            except ValueError as err:
                self.__wrong_src_code_comments.append(str(err))

            review_status.author = self.user_name
            review_status.date = run_history_time

            # False positive and intentional reports are considered as closed
            # reports which is indicated with non-null "fixed_at" date.
            fixed_at = None
            if review_status.status in ['false_positive', 'intentional']:
                # Keep in mind that now this is not handling review status
                # rules, only review status source code comments
                if old_report and old_report.review_status in \
                        ['false_positive', 'intentional']:
                    fixed_at = old_report.review_status_date
                else:
                    fixed_at = run_history_time

            self.__check_report_count()
            self.__add_report(session, run_id, report, report_path_hash,
                              file_path_to_id, review_status, detection_status,
                              detected_at, run_history_time, analysis_info,
                              fixed_at)

            self.__new_report_hashes[report.report_hash] = \
                review_status.status
            self.__already_added_report_hashes.add(report_path_hash)

            LOG.debug("Storing report done. "
                      "path_hash=%s, bug_id/report_hash=%s, source_file=%s",
                      report_path_hash, report.report_hash, report_file_path)

        return True

    def __validate_and_add_report_annotations(
        self,
        session: DBSession,
        report_id: int,
        report_annotation: Dict
    ):
        """
        This function checks the format of the annotations. For example a
        "timestamp" annotation must be in datetime format. If the format
        doesn't match then an exception is thrown. In case of proper format the
        annotation is added to the database.
        """
        ALLOWED_TYPES = {
            "datetime": {
                "func": datetime.fromisoformat,
                "display": "date-time in ISO format"
            },
            "string": {
                "func": str,
                "display": "string"
            }
        }

        ALLOWED_ANNOTATIONS = {
            "timestamp": ALLOWED_TYPES["datetime"],
            "testsuite": ALLOWED_TYPES["string"],
            "testcase": ALLOWED_TYPES["string"]
        }

        for key, value in report_annotation.items():
            try:
                ALLOWED_ANNOTATIONS[key]["func"](value)
                session.add(ReportAnnotations(report_id, key, value))
            except KeyError:
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.REPORT_FORMAT,
                    f"'{key}' is not an allowed report annotation.",
                    ALLOWED_ANNOTATIONS.keys())
            except ValueError:
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.REPORT_FORMAT,
                    f"'{value}' has wrong format. '{key}' annotations must be "
                    f"'{ALLOWED_ANNOTATIONS[key]['display']}'."
                )

    def __get_report_limit_for_product(self):
        with DBSession(self.__config_database) as session:
            product = session.query(Product).get(self.__product.id)
            if product.report_limit:
                self.__report_limit = product.report_limit

    def __check_report_count(self):
        """
        This method comparest the already added report count to the report
        limit, Raises exception if the number of reports is more than the
        that is configured for the product.
        """
        if len(self.__added_reports) >= self.__report_limit:
            LOG.error("The number of reports in the given report folder is " +
                      "larger than the allowed." +
                      f"The limit: {self.__report_limit}!")
            extra_info = [
                "report_limit",
                f"limit:{self.__report_limit}"
            ]
            raise codechecker_api_shared.ttypes.RequestFailed(
                codechecker_api_shared.ttypes.
                ErrorCode.GENERAL,
                "**Report Limit Exceeded** " +
                "This report folder cannot be stored because the number of " +
                "reports in the result folder is too high. Usually noisy " +
                "checkers, generating a lot of reports are not useful and " +
                "it is better to disable them. Run `CodeChecker parse " +
                "<report_folder>` to gain a better understanding on report " +
                "counts. Disable checkers that have generated an excessive " +
                "number of reports and then rerun the analysis to be able " +
                "to store the results on the server. " +
                f"Limit: {self.__report_limit}",
                extra_info)

    def __store_reports(
        self,
        session: DBSession,
        report_dir: str,
        source_root: str,
        run_id: int,
        file_path_to_id: Dict[str, int],
        run_history_time: datetime
    ):
        """ Parse up and store the plist report files. """

        def get_skip_handler(
            report_dir: str
        ) -> skiplist_handler.SkipListHandler:
            """ Get a skip list handler based on the given report directory."""
            skip_file_path = os.path.join(report_dir, 'skip_file')
            if not os.path.exists(skip_file_path):
                return skiplist_handler.SkipListHandler()

            LOG.debug("Pocessing skip file %s", skip_file_path)
            try:
                with open(skip_file_path,
                          encoding="utf-8", errors="ignore") as f:
                    skip_content = f.read()
                    LOG.debug(skip_content)

                    return skiplist_handler.SkipListHandler(skip_content)
            except (IOError, OSError) as err:
                LOG.error("Failed to open skip file: %s", err)
                raise

        # Reset internal data.
        self.__already_added_report_hashes = set()
        self.__new_report_hashes = dict()
        self.__all_report_checkers = set()

        all_reports = session.query(DBReport) \
            .filter(DBReport.run_id == run_id) \
            .all()

        report_to_report_id = defaultdict(list)
        for db_report in all_reports:
            report_to_report_id[db_report.bug_id].append(db_report)

        enabled_checkers: Set[str] = set()
        disabled_checkers: Set[str] = set()

        # Processing analyzer result files.
        processed_result_file_count = 0

        for root_dir_path, _, report_file_paths in os.walk(report_dir):
            LOG.debug("Get reports from '%s' directory", root_dir_path)

            skip_handler = get_skip_handler(root_dir_path)

            review_status_handler = ReviewStatusHandler(source_root)

            review_status_cfg = \
                os.path.join(root_dir_path, 'review_status.yaml')
            if os.path.isfile(review_status_cfg):
                review_status_handler.set_review_status_config(
                    review_status_cfg)

            mip = self.__mips[root_dir_path]
            enabled_checkers.update(mip.enabled_checkers)
            disabled_checkers.update(mip.disabled_checkers)

            for f in report_file_paths:
                if not report_file.is_supported(f):
                    continue

                LOG.debug("Parsing input file '%s'", f)

                report_file_path = os.path.join(root_dir_path, f)
                self.__process_report_file(
                    report_file_path, session, run_id,
                    file_path_to_id, run_history_time,
                    skip_handler, review_status_handler, report_to_report_id)
                processed_result_file_count += 1

        session.flush()

        self.__add_report_context(session, file_path_to_id)
        # Get all relevant review_statuses for the newly stored reports
        # CHHECK: Call self.getReviewStatusRules instead of the below query
        # but before first check the performance
        reports_to_rs_rules = session.query(ReviewStatusRule, DBReport) \
            .join(DBReport, DBReport.bug_id == ReviewStatusRule.bug_hash) \
            .filter(sqlalchemy.and_(
                DBReport.run_id == run_id,
                DBReport.review_status_is_in_source.is_(False),
                ReviewStatusRule.bug_hash.in_(self.__new_report_hashes)))

        # Set the newly stored reports
        for review_status, db_report in reports_to_rs_rules:
            old_report = None
            if db_report.bug_id in report_to_report_id:
                old_report = report_to_report_id[db_report.bug_id][0]
            fixed_at = None
            if review_status.status in ['false_positive', 'intentional']:
                # Keep in mind that now this is not handling review status
                # rules, only review status source code comments
                if old_report and old_report.review_status in \
                        ['false_positive', 'intentional']:
                    fixed_at = old_report.review_status_date
                else:
                    fixed_at = run_history_time

            db_report.review_status = review_status.status
            db_report.review_statuses_author = review_status.author
            db_report.review_status_message = review_status.message
            db_report.review_statuses_date = review_status.date
            db_report.fixed_at = fixed_at
            db_report.review_status_is_in_source = False

        session.flush()

        LOG.info("[%s] Processed %d analyzer result file(s).", self.__name,
                 processed_result_file_count)

        # If a checker was found in a plist file it can not be disabled so we
        # will add this to the enabled checkers list and remove this checker
        # from the disabled checkers list.
        # Also if multiple report directories are stored and a checker was
        # enabled in one report directory but it was disabled in another
        # directory we will mark this checker as enabled.
        enabled_checkers |= self.__all_report_checkers
        disabled_checkers -= self.__all_report_checkers

        reports_to_delete = set()
        for bug_hash, reports in report_to_report_id.items():
            if bug_hash in self.__new_report_hashes:
                reports_to_delete.update([x.id for x in reports])
            else:
                for report in reports:
                    checker_name: str = report.checker.checker_name
                    if checker_name in disabled_checkers:
                        report.detection_status = 'off'
                    elif checker_is_unavailable(checker_name,
                                                enabled_checkers):
                        report.detection_status = 'unavailable'
                    else:
                        report.detection_status = 'resolved'

                    if report.fixed_at is None:
                        report.fixed_at = run_history_time

        if reports_to_delete:
            self.__report_server._removeReports(
                session, list(reports_to_delete))

    def finish_checker_run(
        self,
        session: DBSession,
        run_id: int
    ) -> bool:
        """ Finish the storage of the given run. """
        try:
            LOG.debug("Finishing checker run")
            run = session.query(Run).get(run_id)
            if not run:
                return False

            run.mark_finished()
            run.duration = self.__duration

            return True
        except Exception as ex:
            LOG.error(ex)

        return False

    def store(self) -> int:
        """ Store run results to the server. """
        start_time = time.time()

        # Check constraints of the run.
        self.__check_run_limit()

        with DBSession(self.__Session) as session:
            self.__store_run_lock(session)

        try:
            with TemporaryDirectory(
                dir=self.__context.codechecker_workspace
            ) as zip_dir:
                with LogTask(run_name=self.__name,
                             message="Unzip storage file"):
                    zip_size = unzip(self.__b64zip, zip_dir)

                if zip_size == 0:
                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.
                        ErrorCode.GENERAL,
                        "The received zip file content is empty!")

                LOG.debug("Using unzipped folder '%s'", zip_dir)

                source_root = os.path.join(zip_dir, 'root')
                blame_root = os.path.join(zip_dir, 'blame')
                report_dir = os.path.join(zip_dir, 'reports')
                content_hash_file = os.path.join(
                    zip_dir, 'content_hashes.json')

                filename_to_hash = load_json(content_hash_file, {})

                with LogTask(run_name=self.__name,
                             message="Store source files"):
                    LOG.info("[%s] Storing %d source file(s).", self.__name,
                             len(filename_to_hash.keys()))
                    file_path_to_id = self.__store_source_files(
                        source_root, filename_to_hash)
                    self.__add_blame_info(blame_root, filename_to_hash)

                run_history_time = datetime.now()

                # Parse all metadata information from the report directory.
                with LogTask(run_name=self.__name,
                             message="Parse 'metadata.json's"):
                    for root_dir_path, _, _ in os.walk(report_dir):
                        metadata_file_path = os.path.join(
                            root_dir_path, 'metadata.json')

                        self.__mips[root_dir_path] = \
                            MetadataInfoParser(metadata_file_path)

                with LogTask(run_name=self.__name,
                             message="Store look-up ID for checkers in "
                                     "'metadata.json'"):
                    checkers_in_metadata = {
                        (analyzer, checker)
                        for metadata in self.__mips.values()
                        for analyzer in metadata.analyzers
                        for checker
                        in metadata.checkers.get(analyzer, dict()).keys()}
                    self.__store_checker_identifiers(checkers_in_metadata)

                try:
                    # This session's transaction buffer stores the actual
                    # run data into the database.
                    with DBSession(self.__Session) as session, \
                            RunLocking(session, self.__name):
                        # Actual store operation begins here.
                        run_id, update_run = self.__add_or_update_run(
                            session, run_history_time)

                        with LogTask(run_name=self.__name,
                                     message="Store reports"):
                            self.__store_reports(
                                session, report_dir, source_root, run_id,
                                file_path_to_id, run_history_time)

                        session.commit()
                        self.__load_report_ids_for_reports_with_fake_checkers(
                            session)

                    if self.__reports_with_fake_checkers:
                        with LogTask(run_name=self.__name,
                                     message="Get look-up ID for checkers "
                                     "not present in 'metadata.json'"):
                            additional_checkers = self.__get_faked_checkers()
                            # __store_checker_identifiers() has its own
                            # TRANSACTION!
                            self.__store_checker_identifiers(
                                additional_checkers)

                    with DBSession(self.__Session) as session, \
                            RunLocking(session, self.__name):
                        # The data of the run has been successfully committed
                        # into the database. Deal with post-processing issues
                        # that could only be done after-the-fact.
                        if self.__reports_with_fake_checkers:
                            with LogTask(run_name=self.__name,
                                         message="Fix-up report-to-checker "
                                         "associations"):
                                self.__realise_fake_checkers(session)

                        self.finish_checker_run(session, run_id)
                        session.commit()

                    # If it's a run update, do not increment the number
                    # of runs of the current product.
                    inc_num_of_runs = 1 if not update_run else None

                    self.__report_server._set_run_data_for_curr_product(
                        inc_num_of_runs, run_history_time)

                    runtime = round(time.time() - start_time, 2)
                    zip_size_kb = round(zip_size / 1024)

                    tag_desc = ""
                    if self.__tag:
                        tag_desc = f", under tag '{self.__tag}'"

                    LOG.info("'%s' stored results (%s KB "
                             "/decompressed/) to run '%s' (id: %d) %s in "
                             "%s seconds.", self.user_name,
                             zip_size_kb, self.__name, run_id, tag_desc,
                             runtime)

                    iso_start_time = datetime.fromtimestamp(
                        start_time).isoformat()

                    log_msg = f"{iso_start_time}, " +\
                              f"{runtime}s, " +\
                              f'"{self.__product.name}", ' +\
                              f'"{self.__name}", ' +\
                              f"{zip_size_kb}KB, " +\
                              f"{self.__report_count}, " +\
                              f"{run_id}"

                    STORE_TIME_LOG.info(log_msg)

                    return run_id
                except (sqlalchemy.exc.OperationalError,
                        sqlalchemy.exc.ProgrammingError) as ex:
                    raise codechecker_api_shared.ttypes.RequestFailed(
                        codechecker_api_shared.ttypes.ErrorCode.DATABASE,
                        "Storing reports to the database failed: {0}"
                        .format(ex))
        except Exception as ex:
            LOG.error("Failed to store results: %s", ex)
            import traceback
            traceback.print_exc()
            raise
        finally:
            # In any case if the "try" block's execution began, a run lock must
            # exist, which can now be removed, as storage either completed
            # successfully, or failed in a detectable manner.
            # (If the failure is undetectable, the coded grace period expiry
            # of the lock will allow further store operations to the given
            # run name.)
            with DBSession(self.__Session) as session:
                self.__free_run_lock(session)

            if self.__wrong_src_code_comments:
                raise codechecker_api_shared.ttypes.RequestFailed(
                    codechecker_api_shared.ttypes.ErrorCode.SOURCE_FILE,
                    self.__wrong_src_code_comments)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Thrift enum helper.
"""


from codechecker_api.codeCheckerDBAccess_v6.ttypes import DetectionStatus, \
        ExtendedReportDataType, ReportStatus, ReviewStatus
from codechecker_api.ProductManagement_v6.ttypes import Confidentiality


def detection_status_enum(status):
    if status == 'new':
        return DetectionStatus.NEW
    elif status == 'resolved':
        return DetectionStatus.RESOLVED
    elif status == 'unresolved':
        return DetectionStatus.UNRESOLVED
    elif status == 'reopened':
        return DetectionStatus.REOPENED
    elif status == 'off':
        return DetectionStatus.OFF
    elif status == 'unavailable':
        return DetectionStatus.UNAVAILABLE


def detection_status_str(status):
    if status == DetectionStatus.NEW:
        return 'new'
    elif status == DetectionStatus.RESOLVED:
        return 'resolved'
    elif status == DetectionStatus.UNRESOLVED:
        return 'unresolved'
    elif status == DetectionStatus.REOPENED:
        return 'reopened'
    elif status == DetectionStatus.OFF:
        return 'off'
    elif status == DetectionStatus.UNAVAILABLE:
        return 'unavailable'


def confidentiality_enum(confidentiality: str) -> Confidentiality:
    """
    Converts the given string to confidentiality Thrift enum value.
    """
    if confidentiality == 'CONFIDENTIAL':
        return Confidentiality.CONFIDENTIAL
    elif confidentiality == 'INTERNAL':
        return Confidentiality.INTERNAL
    elif confidentiality == 'OPEN':
        return Confidentiality.OPEN


def confidentiality_str(confidentiality: Confidentiality) -> str:
    """
    Converts the given confidentiality to string.
    """
    if confidentiality == Confidentiality.CONFIDENTIAL:
        return 'CONFIDENTIAL'
    elif confidentiality == Confidentiality.INTERNAL:
        return 'INTERNAL'
    elif confidentiality == Confidentiality.OPEN:
        return 'OPEN'


def review_status_str(status):
    """
    Returns the given review status Thrift enum value.
    """
    if status == ReviewStatus.UNREVIEWED:
        return 'unreviewed'
    elif status == ReviewStatus.CONFIRMED:
        return 'confirmed'
    elif status == ReviewStatus.FALSE_POSITIVE:
        return 'false_positive'
    elif status == ReviewStatus.INTENTIONAL:
        return 'intentional'


def review_status_enum(status):
    """
    Converts the given review status to string.
    """
    if status == 'unreviewed':
        return ReviewStatus.UNREVIEWED
    elif status == 'confirmed':
        return ReviewStatus.CONFIRMED
    elif status == 'false_positive':
        return ReviewStatus.FALSE_POSITIVE
    elif status == 'intentional':
        return ReviewStatus.INTENTIONAL


def report_extended_data_type_str(status):
    """
    Converts the given extended data type to string.
    """
    if status == ExtendedReportDataType.NOTE:
        return 'note'
    elif status == ExtendedReportDataType.MACRO:
        return 'macro'
    elif status == ExtendedReportDataType.FIXIT:
        return 'fixit'


def report_extended_data_type_enum(status):
    """
    Returns the given extended report data Thrift enum value.
    """
    if status == 'note':
        return ExtendedReportDataType.NOTE
    elif status == 'macro':
        return ExtendedReportDataType.MACRO
    elif status == 'fixit':
        return ExtendedReportDataType.FIXIT


def report_status_str(status):
    """
    Returns the given report status Thrift enum value.
    """
    if status == ReportStatus.OUTSTANDING:
        return 'outstanding'
    elif status == ReportStatus.CLOSED:
        return 'closed'


def report_status_enum(status):
    """
    Converts the given report status to string.
    """
    if status == 'outstanding':
        return ReportStatus.OUTSTANDING
    elif status == 'closed':
        return ReportStatus.CLOSED



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Result handler for Gcc.
"""
from typing import Optional
from pathlib import Path
import shutil
import os

from codechecker_report_converter.report.parser.base import AnalyzerInfo
from codechecker_report_converter.analyzers.gcc.analyzer_result import \
    AnalyzerResult
from codechecker_report_converter.report import report_file
from codechecker_report_converter.report.hash import get_report_hash, HashType

from codechecker_common.logger import get_logger
from codechecker_common.skiplist_handler import SkipListHandlers
from codechecker_common.review_status_handler import ReviewStatusHandler

from ..result_handler_base import ResultHandler

LOG = get_logger('analyzer.gcc')


def actual_name_to_codechecker_name(actual_name: str):
    assert actual_name.startswith('-Wanalyzer')
    return actual_name.replace("-Wanalyzer", "gcc")


def codechecker_name_to_actual_name(codechecker_name: str):
    assert codechecker_name.startswith('gcc')
    return codechecker_name.replace("gcc", "-Wanalyzer")


def codechecker_name_to_actual_name_disabled(codechecker_name: str):
    assert codechecker_name.startswith('gcc')
    return codechecker_name.replace("gcc", "-Wno-analyzer")


class GccResultHandler(ResultHandler):
    """
    Create analyzer result file for Gcc output.
    """

    def __init__(self, *args, **kwargs):
        self.analyzer_info = AnalyzerInfo(name=AnalyzerResult.TOOL_NAME)
        self.gcc_analyzer_result = AnalyzerResult()

        super(GccResultHandler, self).__init__(*args, **kwargs)

    def postprocess_result(
        self,
        skip_handlers: Optional[SkipListHandlers],
        rs_handler: Optional[ReviewStatusHandler]
    ):
        """
        Generate analyzer result output file which can be parsed and stored
        into the database.
        """
        LOG.debug_analyzer(self.analyzer_stdout)
        gcc_stderr = self.analyzer_stderr
        assert gcc_stderr, "Even in the event of no reported diagnostics, " \
                           "stderr mustn't be empty!"

        # report-converter needs a file to parse, let's dump the content of
        # stderr to one.
        gcc_out_folder = Path(self.workspace, "gcc")
        gcc_out_folder.mkdir(exist_ok=True)
        gcc_dest_file_name = \
            str(Path(gcc_out_folder,
                     os.path.basename(self.analyzed_source_file) +
                     self.buildaction_hash + ".sarif"))

        with open(gcc_dest_file_name, 'w') as f:
            f.write(gcc_stderr)
        assert os.path.exists(gcc_dest_file_name)

        reports = report_file.get_reports(
            gcc_dest_file_name, self.checker_labels,
            source_dir_path=self.source_dir_path)

        # FIXME: We absolutely want to support gcc compiler warnings
        # eventually (which don't start with '-Wanalyzer'), but we should
        # probably list them in the label files as well, etc.
        reports = \
            [r for r in reports if not r.skip(skip_handlers) and
             r.checker_name.startswith("-Wanalyzer")]

        # If we were to leave sarif files in the repoort directory, we would
        # unintentionally parse them, so we rename them.
        try:
            shutil.move(gcc_dest_file_name, gcc_dest_file_name + '.bak')
        except(OSError) as e:
            LOG.error(e)

        for report in reports:
            report.checker_name = \
                actual_name_to_codechecker_name(report.checker_name)

        hash_type = HashType.PATH_SENSITIVE
        if self.report_hash_type == 'context-free-v2':
            hash_type = HashType.CONTEXT_FREE
        elif self.report_hash_type == 'diagnostic-message':
            hash_type = HashType.DIAGNOSTIC_MESSAGE

        for report in reports:
            report.report_hash = get_report_hash(report, hash_type)

        if rs_handler:
            reports = [r for r in reports if not rs_handler.should_ignore(r)]

        report_file.create(
            self.analyzer_result_file, reports, self.checker_labels,
            self.analyzer_info)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Handles the management of authentication sessions on the server's side.
"""

import hashlib
import json
import os
import re
import uuid

from datetime import datetime
from typing import Optional

from codechecker_common.compatibility.multiprocessing import cpu_count
from codechecker_common.logger import get_logger
from codechecker_common.util import load_json

from codechecker_web.shared.env import check_file_owner_rw
from codechecker_web.shared.version import SESSION_COOKIE_NAME as _SCN

from .database.config_db_model import Session as SessionRecord
from .database.config_db_model import SystemPermission
from .permissions import SUPERUSER


UNSUPPORTED_METHODS = []

try:
    from .auth import cc_ldap
except ImportError:
    UNSUPPORTED_METHODS.append('ldap')

try:
    from .auth import cc_pam
except ImportError:
    UNSUPPORTED_METHODS.append('pam')


LOG = get_logger("server")
SESSION_COOKIE_NAME = _SCN


def generate_session_token():
    """
    Returns a random session token.
    """
    return uuid.UUID(bytes=os.urandom(16)).hex


def get_worker_processes(scfg_dict):
    """
    Return number of worker processes from the config dictionary.

    Return 'worker_processes' field from the config dictionary or returns the
    default value if this field is not set or the value is negative.
    """
    default = cpu_count()
    worker_processes = scfg_dict.get('worker_processes', default)

    if worker_processes < 0:
        LOG.warning("Number of worker processes can not be negative! Default "
                    "value will be used: %s", default)
        worker_processes = default

    return worker_processes


class _Session:
    """A session for an authenticated, privileged client connection."""

    def __init__(self, token, username, groups,
                 session_lifetime, refresh_time, is_root=False, database=None,
                 last_access=None, can_expire=True):

        self.token = token
        self.user = username
        self.groups = groups

        self.session_lifetime = session_lifetime
        self.refresh_time = refresh_time if refresh_time else None
        self.__root = is_root
        self.__database = database
        self.__can_expire = can_expire
        self.last_access = last_access if last_access else datetime.now()

    @property
    def is_root(self):
        """Returns whether or not the Session was created with the master
        superuser (root) credentials."""
        return self.__root

    @property
    def is_refresh_time_expire(self):
        """
        Returns if the refresh time of the session is expired.
        """
        if not self.refresh_time:
            return True

        return (datetime.now() - self.last_access).total_seconds() > \
            self.refresh_time

    @property
    def is_alive(self):
        """
        Returns if the session is alive and usable, that is, within its
        lifetime.
        """
        if not self.__can_expire:
            return True

        return (datetime.now() - self.last_access).total_seconds() <= \
            self.session_lifetime

    def revalidate(self):
        """
        A session is only revalidated if it has yet to exceed its
        lifetime. After a session hasn't been used for this interval,
        it can NOT be resurrected at all --- the user needs to log in
        to a brand-new session.
        """

        if not self.is_alive:
            return

        if self.__database and self.is_refresh_time_expire:
            self.last_access = datetime.now()

            # Update the timestamp in the database for the session's last
            # access.
            transaction = None
            try:
                transaction = self.__database()
                record = transaction.query(SessionRecord) \
                    .filter(SessionRecord.user_name == self.user) \
                    .filter(SessionRecord.token == self.token) \
                    .limit(1).one_or_none()

                if record:
                    record.last_access = self.last_access
                    transaction.commit()
            except Exception as e:
                LOG.warning("Couldn't update usage timestamp of %s",
                            self.token)
                LOG.warning(str(e))
            finally:
                if transaction:
                    transaction.close()


class SessionManager:
    """
    Provides the functionality required to handle user authentication on a
    CodeChecker server.
    """

    def __init__(self, configuration_file, root_sha, force_auth=False):
        """
        Initialise a new Session Manager on the server.

        :param configuration_file: The configuration file to read
            authentication backends from.
        :param root_sha: The SHA-256 hash of the root user's authentication.
        :param force_auth: If True, the manager will be enabled even if the
            configuration file disables authentication.
        """
        self.__database_connection = None
        self.__logins_since_prune = 0
        self.__sessions = []
        self.__configuration_file = configuration_file

        scfg_dict = self.__get_config_dict()

        # FIXME: Refactor this. This is irrelevant to authentication config,
        # so it should NOT be handled by session_manager. A separate config
        # handler for the server's stuff should be created, that can properly
        # instantiate SessionManager with the found configuration.
        self.__worker_processes = get_worker_processes(scfg_dict)
        self.__max_run_count = scfg_dict.get('max_run_count', None)
        self.__store_config = scfg_dict.get('store', {})
        self.__keepalive_config = scfg_dict.get('keepalive', {})
        self.__auth_config = scfg_dict['authentication']

        if force_auth:
            LOG.debug("Authentication was force-enabled.")
            self.__auth_config['enabled'] = True

        if 'soft_expire' in self.__auth_config:
            LOG.debug("Found deprecated argument 'soft_expire' in "
                      "server_config.authentication.")

        self.__refresh_time = self.__auth_config['refresh_time'] \
            if 'refresh_time' in self.__auth_config else None

        # Save the root SHA into the configuration (but only in memory!)
        self.__auth_config['method_root'] = root_sha

        self.__regex_groups_enabled = False

        # Pre-compile the regular expressions of 'regex_groups'
        if 'regex_groups' in self.__auth_config:
            self.__regex_groups_enabled = self.__auth_config['regex_groups'] \
                                              .get('enabled', False)

            regex_groups = self.__auth_config['regex_groups'] \
                               .get('groups', [])
            d = dict()
            for group_name, regex_list in regex_groups.items():
                d[group_name] = [re.compile(r) for r in regex_list]
            self.__group_regexes_compiled = d

        # If no methods are configured as enabled, disable authentication.
        if scfg_dict['authentication'].get('enabled'):
            found_auth_method = False

            if 'method_dictionary' in self.__auth_config and \
                    self.__auth_config['method_dictionary'].get('enabled'):
                found_auth_method = True

            if 'method_ldap' in self.__auth_config and \
                    self.__auth_config['method_ldap'].get('enabled'):
                if 'ldap' not in UNSUPPORTED_METHODS:
                    found_auth_method = True
                else:
                    LOG.warning("LDAP authentication was enabled but "
                                "prerequisites are NOT installed on the system"
                                "... Disabling LDAP authentication.")
                    self.__auth_config['method_ldap']['enabled'] = False

            if 'method_pam' in self.__auth_config and \
                    self.__auth_config['method_pam'].get('enabled'):
                if 'pam' not in UNSUPPORTED_METHODS:
                    found_auth_method = True
                else:
                    LOG.warning("PAM authentication was enabled but "
                                "prerequisites are NOT installed on the system"
                                "... Disabling PAM authentication.")
                    self.__auth_config['method_pam']['enabled'] = False

            if not found_auth_method:
                if force_auth:
                    LOG.warning("Authentication was manually enabled, but no "
                                "valid authentication backends are "
                                "configured... The server will only allow "
                                "the master superuser (root) access.")
                else:
                    LOG.warning("Authentication is enabled but no valid "
                                "authentication backends are configured... "
                                "Falling back to no authentication.")
                    self.__auth_config['enabled'] = False

    def __get_config_dict(self):
        """
        Get server config information from the configuration file. Raise
        ValueError if the configuration file is invalid.
        """
        LOG.debug(self.__configuration_file)
        cfg_dict = load_json(self.__configuration_file, {})
        if cfg_dict != {}:
            check_file_owner_rw(self.__configuration_file)
        else:
            # If the configuration dict is empty, it means a JSON couldn't
            # have been parsed from it.
            raise ValueError("Server configuration file was invalid, or "
                             "empty.")
        return cfg_dict

    def reload_config(self):
        LOG.info("Reload server configuration file...")
        try:
            cfg_dict = self.__get_config_dict()

            prev_max_run_count = self.__max_run_count
            new_max_run_count = cfg_dict.get('max_run_count', None)
            if prev_max_run_count != new_max_run_count:
                self.__max_run_count = new_max_run_count
                LOG.debug("Changed 'max_run_count' value from %s to %s",
                          prev_max_run_count, new_max_run_count)

            prev_store_config = json.dumps(self.__store_config, sort_keys=True,
                                           indent=2)
            new_store_config_val = cfg_dict.get('store', {})
            new_store_config = json.dumps(new_store_config_val, sort_keys=True,
                                          indent=2)
            if prev_store_config != new_store_config:
                self.__store_config = new_store_config_val
                LOG.debug("Updating 'store' config from %s to %s",
                          prev_store_config, new_store_config)

            update_sessions = False
            auth_fields_to_update = ['session_lifetime', 'refresh_time',
                                     'logins_until_cleanup']
            for field in auth_fields_to_update:
                if field in self.__auth_config:
                    prev_value = self.__auth_config[field]
                    new_value = cfg_dict['authentication'].get(field, 0)
                    if prev_value != new_value:
                        self.__auth_config[field] = new_value
                        LOG.debug("Changed '%s' value from %s to %s",
                                  field, prev_value, new_value)
                        update_sessions = True

            if update_sessions:
                # Update configuration options of the already existing
                # sessions.
                for session in self.__sessions:
                    session.session_lifetime = \
                        self.__auth_config['session_lifetime']
                    session.refresh_time = self.__auth_config['refresh_time']

            LOG.info("Done.")
        except ValueError as ex:
            LOG.error("Couldn't reload server configuration file")
            LOG.error(str(ex))

    @property
    def is_enabled(self):
        return self.__auth_config.get('enabled')

    @property
    def worker_processes(self):
        return self.__worker_processes

    def get_realm(self):
        return {
            "realm": self.__auth_config.get('realm_name'),
            "error": self.__auth_config.get('realm_error')
        }

    @property
    def default_superuser_name(self) -> Optional[str]:
        """ Get default superuser name. """
        root = self.__auth_config['method_root'].split(":")

        # Previously the root file doesn't contain the user name. In this case
        # we will return with no user name.
        if len(root) <= 1:
            return None

        return root[0]

    def set_database_connection(self, connection):
        """
        Set the instance's database connection to use in fetching
        database-stored sessions to the given connection.

        Use None as connection's value to unset the database.
        """
        self.__database_connection = connection

    def __handle_validation(self, auth_string):
        """
        Validate an oncoming authorization request
        against some authority controller.

        Returns False if no validation was done, or a validation object
        if the user was successfully authenticated.

        This validation object contains two keys: username and groups.
        """
        validation = self.__try_auth_root(auth_string) \
            or self.__try_auth_dictionary(auth_string) \
            or self.__try_auth_pam(auth_string) \
            or self.__try_auth_ldap(auth_string)
        if not validation:
            return False

        # If a validation method is enabled and regex_groups is enabled too,
        # we will extend the 'groups'.
        extra_groups = self.__try_regex_groups(validation['username'])
        if extra_groups:
            already_groups = set(validation['groups'])
            validation['groups'] = list(already_groups | extra_groups)

        LOG.debug('User validation details: %s', str(validation))
        return validation

    def __is_method_enabled(self, method):
        return method not in UNSUPPORTED_METHODS and \
            'method_' + method in self.__auth_config and \
            self.__auth_config['method_' + method].get('enabled')

    def __try_auth_root(self, auth_string):
        """
        Try to authenticate the user against the root username:password's hash.
        """
        user_name = SessionManager.get_user_name(auth_string)
        sha = hashlib.sha256(auth_string.encode('utf8')).hexdigest()

        if f"{user_name}:{sha}" == self.__auth_config['method_root']:
            return {
                'username': SessionManager.get_user_name(auth_string),
                'groups': [],
                'root': True
            }

        return False

    def __try_auth_token(self, auth_string):
        if not self.__database_connection:
            return None

        user_name, token = auth_string.split(':', 1)

        transaction = None
        try:
            # Try the database, if it is connected.
            transaction = self.__database_connection()
            auth_session = transaction.query(SessionRecord.token) \
                .filter(SessionRecord.user_name == user_name) \
                .filter(SessionRecord.token == token) \
                .filter(SessionRecord.can_expire.is_(False)) \
                .limit(1).one_or_none()

            if not auth_session:
                return False

            return auth_session
        except Exception as e:
            LOG.error("Couldn't check login in the database: ")
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

    def __try_auth_dictionary(self, auth_string):
        """
        Try to authenticate the user against the hardcoded credential list.

        Returns a validation object if successful, which contains the users'
        groups.
        """
        method_config = self.__auth_config.get('method_dictionary')
        if not method_config:
            return False

        valid = self.__is_method_enabled('dictionary') and \
            auth_string in method_config.get('auths')
        if not valid:
            return False

        username = SessionManager.get_user_name(auth_string)
        group_list = method_config['groups'][username] if \
            'groups' in method_config and \
            username in method_config['groups'] else []

        return {
            'username': username,
            'groups': group_list
        }

    def __try_auth_pam(self, auth_string):
        """
        Try to authenticate user based on the PAM configuration.
        """
        if self.__is_method_enabled('pam'):
            username, password = auth_string.split(':', 1)
            if cc_pam.auth_user(self.__auth_config['method_pam'],
                                username, password):
                # PAM does not hold a group membership list we can reliably
                # query.
                return {'username': username}

        return False

    def __try_auth_ldap(self, auth_string):
        """
        Try to authenticate user to all the configured authorities.
        """
        if self.__is_method_enabled('ldap'):
            username, password = auth_string.split(':', 1)

            ldap_authorities = self.__auth_config['method_ldap'] \
                .get('authorities')
            for ldap_conf in ldap_authorities:
                if cc_ldap.auth_user(ldap_conf, username, password):
                    groups = cc_ldap.get_groups(ldap_conf, username, password)
                    self.__update_groups(username, groups)
                    return {'username': username, 'groups': groups}

        return False

    def __update_groups(self, user_name, groups):
        """
        Updates group field of the users tokens.
        """
        if not self.__database_connection:
            return None

        transaction = None
        try:
            # Try the database, if it is connected.
            transaction = self.__database_connection()
            transaction.query(SessionRecord) \
                .filter(SessionRecord.user_name == user_name) \
                .update({SessionRecord.groups: ';'.join(groups)})
            transaction.commit()
            return True
        except Exception as e:
            LOG.error("Couldn't check login in the database: ")
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

        return False

    def __try_regex_groups(self, username):
        """
        Return a set of groups that the user belongs to, depending on whether
        the username matches the regular expression of the group.

        """
        if not self.__regex_groups_enabled:
            return set()

        matching_groups = set()
        for group_name, regex_list in self.__group_regexes_compiled.items():
            for r in regex_list:
                if re.search(r, username):
                    matching_groups.add(group_name)

        return matching_groups

    @staticmethod
    def get_user_name(auth_string):
        return auth_string.split(':')[0]

    def get_db_auth_session_tokens(self, user_name):
        """
        Get authentication session token from the database for the given user.
        """
        if not self.__database_connection:
            return None

        transaction = None
        try:
            # Try the database, if it is connected.
            transaction = self.__database_connection()
            session_tokens = transaction.query(SessionRecord) \
                .filter(SessionRecord.user_name == user_name) \
                .filter(SessionRecord.can_expire.is_(True)) \
                .all()
            return session_tokens
        except Exception as e:
            LOG.error("Couldn't check login in the database: ")
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

        return None

    def __is_root_user(self, user_name):
        """ Return True if the given user has system permissions. """
        if self.__auth_config['method_root'].split(":")[0] == user_name:
            return True

        transaction = None
        try:
            # Try the database, if it is connected.
            transaction = self.__database_connection()
            system_permission = transaction.query(SystemPermission) \
                .filter(SystemPermission.name == user_name) \
                .filter(SystemPermission.permission == SUPERUSER.name) \
                .limit(1).one_or_none()
            return True if system_permission else False
        except Exception as e:
            LOG.error("Couldn't get system permission from database: ")
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

        return False

    def __create_local_session(self, token, user_name, groups, is_root,
                               last_access=None, can_expire=True):
        """
        Returns a new local session object initalized by the given parameters.
        """
        if not is_root:
            is_root = self.__is_root_user(user_name)

        return _Session(
            token, user_name, groups,
            self.__auth_config['session_lifetime'],
            self.__refresh_time, is_root, self.__database_connection,
            last_access, can_expire)

    def create_session(self, auth_string):
        """ Creates a new session for the given auth-string. """
        if not self.__auth_config['enabled']:
            return None

        # Perform cleanup of session memory, if neccessary.
        self.__logins_since_prune += 1
        if self.__logins_since_prune >= \
                self.__auth_config['logins_until_cleanup']:
            self.__cleanup_sessions()

        # Try authenticate user with personal access token.
        auth_token = self.__try_auth_token(auth_string)
        if auth_token:
            local_session = self.__get_local_session_from_db(auth_token.token)
            local_session.revalidate()
            self.__sessions.append(local_session)
            return local_session

        # Try to authenticate user with different authentication methods.
        validation = self.__handle_validation(auth_string)
        if not validation:
            return False

        # Generate a new token and create a local session.
        token = generate_session_token()
        user_name = validation.get('username')
        groups = validation.get('groups', [])
        is_root = validation.get('root', False)

        local_session = self.__create_local_session(token, user_name,
                                                    groups, is_root)
        self.__sessions.append(local_session)

        # Store the session in the database.
        transaction = None
        if self.__database_connection:
            try:
                transaction = self.__database_connection()
                record = SessionRecord(token, user_name,
                                       ';'.join(groups))
                transaction.add(record)
                transaction.commit()
            except Exception as e:
                LOG.error("Couldn't store or update login record in "
                          "database:")
                LOG.error(str(e))
            finally:
                if transaction:
                    transaction.close()

        return local_session

    def get_max_run_count(self):
        """
        Returns the maximum storable run count. If the value is None it means
        we can upload unlimited number of runs.
        """
        return self.__max_run_count

    def get_analysis_statistics_dir(self):
        """
        Get directory where the compressed analysis statistics files should be
        stored. If the value is None it means we do not want to store
        analysis statistics information on the server.
        """

        return self.__store_config.get('analysis_statistics_dir')

    def get_failure_zip_size(self):
        """
        Maximum size of the collected failed zips which can be store on the
        server.
        """
        limit = self.__store_config.get('limit', {})
        return limit.get('failure_zip_size')

    def get_compilation_database_size(self):
        """
        Limit of the compilation database file size.
        """
        limit = self.__store_config.get('limit', {})
        return limit.get('compilation_database_size')

    def is_keepalive_enabled(self):
        """
        True if the keepalive functionality is explicitly enabled, otherwise it
        will return False.
        """
        return self.__keepalive_config.get('enabled')

    def get_keepalive_idle(self):
        """ Get keepalive idle time. """
        return self.__keepalive_config.get('idle')

    def get_keepalive_interval(self):
        """ Get keepalive interval time. """
        return self.__keepalive_config.get('interval')

    def get_keepalive_max_probe(self):
        """ Get keepalive max probe count. """
        return self.__keepalive_config.get('max_probe')

    def __get_local_session_from_db(self, token):
        """
        Creates a local session if a valid session token can be found in the
        database.
        """

        if not self.__database_connection:
            return

        transaction = None
        try:
            transaction = self.__database_connection()
            db_record = transaction.query(SessionRecord) \
                .filter(SessionRecord.token == token) \
                .limit(1).one_or_none()

            if db_record:
                user_name = db_record.user_name
                is_root = self.__is_root_user(user_name)

                groups = db_record.groups.split(';') \
                    if db_record.groups else []

                return self.__create_local_session(token, user_name,
                                                   groups,
                                                   is_root,
                                                   db_record.last_access,
                                                   db_record.can_expire)
        except Exception as e:
            LOG.error("Couldn't check login in the database: ")
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

    def get_session(self, token):
        """
        Retrieves the session for the given session cookie token from the
        server's memory backend, if such session exists or creates and returns
        a new one if the token exists in the database.

        :returns: The session object if one was found. None if authentication
        is not enabled, or if the cookie is not valid anymore.
        """

        if not self.is_enabled:
            return None

        for sess in self.__sessions:
            if sess.is_alive and sess.token == token:
                # If the session is alive but the should be re-validated.
                if sess.is_refresh_time_expire:
                    sess.revalidate()
                return sess

        # Try to get a local session from the database.
        local_session = self.__get_local_session_from_db(token)
        if local_session and local_session.is_alive:
            self.__sessions.append(local_session)
            if local_session.is_refresh_time_expire:
                local_session.revalidate()
            return local_session

        self.invalidate(token)

        return None

    def invalidate_local_session(self, token):
        """
        Remove a user's previous session from the local in memory store.
        """
        for session in self.__sessions[:]:
            if session.token == token:
                self.__sessions.remove(session)
                return True
        return False

    def invalidate(self, token):
        """
        Remove a user's previous session from local in memory and the database
        store.
        """
        transaction = None
        try:
            self.invalidate_local_session(token)

            transaction = self.__database_connection() \
                if self.__database_connection else None

            # Remove sessions from the database.
            if transaction:
                transaction.query(SessionRecord) \
                    .filter(SessionRecord.token == token) \
                    .filter(SessionRecord.can_expire.is_(True)) \
                    .delete()
                transaction.commit()

            return True
        except Exception as e:
            LOG.error("Couldn't invalidate session for token %s", token)
            LOG.error(str(e))
        finally:
            if transaction:
                transaction.close()

        return False

    def __cleanup_sessions(self):
        self.__logins_since_prune = 0

        for s in self.__sessions:
            if s.is_refresh_time_expire:
                self.invalidate_local_session(s.token)

        for s in self.__sessions:
            if not s.is_alive:
                self.invalidate(s.token)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Defines the routing rules for the CodeChecker server.
"""


import re
from urllib.parse import urlparse

from codechecker_web.shared.version import SUPPORTED_VERSIONS

# A list of top-level path elements under the webserver root
# which should not be considered as a product route.
NON_PRODUCT_ENDPOINTS = ['index.html',
                         'images',
                         'docs',
                         'live',
                         'ready']

# A list of top-level path elements in requests (such as Thrift endpoints)
# which should not be considered as a product route.
NON_PRODUCT_ENDPOINTS += ['Authentication',
                          'Products',
                          'CodeCheckerService']


# A list of top-level path elements under the webserver root which should
# be protected by authentication requirement when accessing the server.
PROTECTED_ENTRY_POINTS = ['',  # Empty string in a request is 'index.html'.
                          'index.html']


def is_valid_product_endpoint(uripart):
    """
    Returns whether or not the given URI part is to be considered a valid
    product name.
    """
    # FIXME: Endpoint "all" should be disallowed, as commit
    # fd59927013d5482ff10e80994511971770753d0c in Dec 2017 added the ability
    # for "CodeChecker server" to specify "--db-status all" and
    # "--db-upgrade-schema all" for the case where *every* product needs to
    # be checked/upgraded, essentially blocking the ability to status-check
    # or schema migrate the product at the endpoint literal "all".

    # There are some forbidden keywords.
    if uripart in NON_PRODUCT_ENDPOINTS:
        return False

    pattern = r'^[A-Za-z0-9_\-]+$'
    if not re.match(pattern, uripart):
        return False

    return True


def is_supported_version(version):
    """
    Returns whether or not the given version tag is supported by the current
    build. A version is supported if its MAJOR version is supported, and if
    its MINOR version is at most the highest minor version accepted by the
    server.

    If supported, returns the major and minor version as a tuple.
    """

    version = version.lstrip('v')
    version_parts = version.split('.')

    # We don't care if accidentally the version tag contains a revision number.
    major, minor = int(version_parts[0]), int(version_parts[1])
    if major in SUPPORTED_VERSIONS and minor <= SUPPORTED_VERSIONS[major]:
        return major, minor

    return False


def split_client_GET_request(path):
    """
    Split the given request URI to its parts relevant to the server.

    Returns the product endpoint and the "remainder" of the request path
    as a tuple of 2.
    """

    # A standard GET request from a browser looks like:
    # http://localhost:8001/[product-name]/#{request-parts}
    # where the parts are, e.g.: run=[run_id]&report=[report_id]

    parsed_path = urlparse(path).path
    split_path = parsed_path.split('/', 2)

    endpoint_part = split_path[1] if len(split_path) > 1 else None
    if endpoint_part and is_valid_product_endpoint(endpoint_part):
        remainder = split_path[2] if len(split_path) == 3 else ''
        return endpoint_part, remainder
    else:
        # The request wasn't pointing to a valid product endpoint.
        return None, parsed_path.lstrip('/')


def split_client_POST_request(path):
    """
    Split the given request URI to its parts relevant to the server.

    Returns the product endpoint, the API version and the API service endpoint
    as a tuple of 3.
    """

    # A standard POST request from an API client looks like:
    # http://localhost:8001/[product-name]/<API version>/<API service>
    # where specifying the product name is optional.

    split_path = urlparse(path).path.split('/', 3)

    endpoint_part = split_path[1]
    if is_valid_product_endpoint(split_path[1]):
        version_tag = split_path[2].lstrip('v')
        remainder = split_path[3]

        return endpoint_part, version_tag, remainder
    elif split_path[1].startswith('v'):
        # Request came through without a valid product URL endpoint to
        # possibly the main server.
        version_tag = split_path[1].lstrip('v')
        remainder = split_path[2]

        return None, version_tag, remainder

    return None, None, None


def is_protected_GET_entrypoint(path):
    """
    Returns if the given GET request's PATH enters the server through an
    entry point which is considered protected by authentication requirements.
    """
    return path in PROTECTED_ENTRY_POINTS



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
"""


import glob
import os
import shlex
import shutil
import signal
import sys
import time
import traceback
import zipfile

from threading import Timer

import multiprocess
import psutil

from codechecker_common.logger import get_logger
from codechecker_common.review_status_handler import ReviewStatusHandler

from codechecker_statistics_collector.collectors.special_return_value import \
    SpecialReturnValueCollector

from . import gcc_toolchain

from .analyzers import analyzer_types
from .analyzers.config_handler import CheckerState
from .analyzers.clangsa.analyzer import ClangSA

LOG = get_logger('analyzer')


def print_analyzer_statistic_summary(metadata_analyzers, status, msg=None):
    """
    Print analyzer statistic summary for the given status code with the given
    section heading message.
    """
    has_status = False
    for _, analyzer in metadata_analyzers.items():
        if analyzer.get('analyzer_statistics', {}).get(status):
            has_status = True
            break

    if has_status and msg:
        LOG.info(msg)

    for analyzer_type, analyzer in metadata_analyzers.items():
        res = analyzer.get('analyzer_statistics', {}).get(status)
        if res:
            LOG.info("  %s: %s", analyzer_type, res)


def worker_result_handler(results, metadata_tool, output_path):
    """ Print the analysis summary. """
    skipped_num = 0
    reanalyzed_num = 0
    metadata_analyzers = metadata_tool['analyzers']
    for res, skipped, reanalyzed, analyzer_type, _, sources in results:
        statistics = metadata_analyzers[analyzer_type]['analyzer_statistics']
        if skipped:
            skipped_num += 1
        else:
            if reanalyzed:
                reanalyzed_num += 1

            if res == 0:
                statistics['successful'] += 1
                statistics['successful_sources'].append(sources)
            else:
                statistics['failed'] += 1
                statistics['failed_sources'].append(sources)

    LOG.info("----==== Summary ====----")
    print_analyzer_statistic_summary(metadata_analyzers,
                                     'successful',
                                     'Successfully analyzed')

    print_analyzer_statistic_summary(metadata_analyzers,
                                     'failed',
                                     'Failed to analyze')

    if reanalyzed_num:
        LOG.info("Reanalyzed compilation commands: %d", reanalyzed_num)
    if skipped_num:
        LOG.info("Skipped compilation commands: %d", skipped_num)

    metadata_tool['skipped'] = skipped_num

    # check() created the result .plist files and additional, per-analysis
    # meta information in forms of .plist.source files.
    # We now soak these files into the metadata dict, as they are not needed
    # as loose files on the disk... but synchronizing LARGE dicts between
    # threads would be more error prone.
    source_map = {}
    for f in glob.glob(os.path.join(output_path, "*.source")):
        with open(f, 'r', encoding="utf-8", errors="ignore") as sfile:
            source_map[f[:-7]] = sfile.read().strip()
        os.remove(f)

    for f in glob.glob(os.path.join(output_path, 'failed', "*.error")):
        err_file, _ = os.path.splitext(f)
        plist_file = os.path.basename(err_file) + ".plist"
        plist_file = os.path.join(output_path, plist_file)
        metadata_tool['result_source_files'].pop(plist_file, None)

    metadata_tool['result_source_files'].update(source_map)


# Progress reporting.
progress_checked_num = None
progress_actions = None


def init_worker(checked_num, action_num):
    global progress_checked_num, progress_actions
    progress_checked_num = checked_num
    progress_actions = action_num


def save_output(base_file_name, out, err):
    try:
        if out:
            with open(base_file_name + ".stdout.txt", 'w',
                      encoding="utf-8", errors="ignore") as outf:
                outf.write(out)

        if err:
            with open(base_file_name + ".stderr.txt", 'w',
                      encoding="utf-8", errors="ignore") as outf:
                outf.write(err)
    except IOError as ioerr:
        LOG.debug("Failed to save analyzer output")
        LOG.debug(ioerr)


def save_metadata(result_file, analyzer_result_file, analyzed_source_file):
    """
    Save some extra information next to the plist, .source
    acting as an extra metadata file.
    """
    with open(result_file + ".source", 'w',
              encoding="utf-8", errors="ignore") as orig:
        orig.write(analyzed_source_file.replace(r'\ ', ' ') + "\n")

    if os.path.exists(analyzer_result_file) and \
            not os.path.exists(result_file):
        os.rename(analyzer_result_file, result_file)


def is_ctu_active(source_analyzer):
    """
    Check if CTU analysis is active for Clang Static Analyzer.
    """
    if not isinstance(source_analyzer, ClangSA):
        return False

    return source_analyzer.is_ctu_available() and \
        source_analyzer.is_ctu_enabled()


def prepare_check(action, analyzer_config, output_dir,
                  skip_handlers, statistics_data, disable_ctu=False):
    """ Construct the source analyzer and result handler. """
    # Create a source analyzer.
    source_analyzer = \
        analyzer_types.construct_analyzer(action,
                                          analyzer_config)

    if disable_ctu:
        # WARNING! can be called only on ClangSA
        # Needs to be called before construct_analyzer_cmd
        source_analyzer.disable_ctu()

    if action.analyzer_type == ClangSA.ANALYZER_NAME and \
       statistics_data:
        # WARNING! Statistical checkers are only supported by Clang
        # Static Analyzer right now.
        stats_dir = statistics_data['stats_out_dir']

        # WARNING Because both statistical checkers use the same config
        # directory it is enough to add it only once. This might change later.
        # Configuration arguments should be added before the checkers are
        # enabled.
        stats_cfg = \
            SpecialReturnValueCollector.checker_analyze_cfg(stats_dir)

        if os.path.exists(SpecialReturnValueCollector.stats_file(stats_dir)):
            source_analyzer.add_checker_config(stats_cfg)
        else:
            LOG.debug('No checker statistics file was found for %s',
                      SpecialReturnValueCollector.checker_analyze)

    # Source is the currently analyzed source file
    # there can be more in one buildaction.
    source_analyzer.source_file = action.source

    # The result handler for analysis is an empty result handler
    # which only returns metadata, but can't process the results.
    rh = source_analyzer.construct_result_handler(action,
                                                  output_dir,
                                                  skip_handlers)

    # NOTICE!
    # The currently analyzed source file needs to be set before the
    # analyzer command is constructed.
    # The analyzer output file is based on the currently
    # analyzed source.
    rh.analyzed_source_file = action.source
    return source_analyzer, rh


def handle_success(
    rh, result_file, result_base, skip_handlers,
    rs_handler: ReviewStatusHandler,
    capture_analysis_output, success_dir
):
    """
    Result postprocessing is required if the analysis was
    successful (mainly clang tidy output conversion is done).

    Skipping reports for header files is done here too.
    """
    if capture_analysis_output:
        save_output(os.path.join(success_dir, result_base),
                    rh.analyzer_stdout, rh.analyzer_stderr)

    rh.postprocess_result(skip_handlers, rs_handler)

    # Generated reports will be handled separately at store.

    save_metadata(result_file, rh.analyzer_result_file,
                  rh.analyzed_source_file)


def handle_reproducer(source_analyzer, rh, zip_file, actions_map):
    """
    If the analysis fails a debug zip is packed together which contains
    build, analysis information and source files to be able to
    reproduce the failed analysis.
    """
    other_files = set()
    action = rh.buildaction

    try:
        LOG.debug("Fetching other dependent files from analyzer output...")
        other_files.update(
            source_analyzer.get_analyzer_mentioned_files(
                rh.analyzer_stdout))

        other_files.update(
            source_analyzer.get_analyzer_mentioned_files(
                rh.analyzer_stderr))
    except Exception as ex:
        LOG.debug("Couldn't generate list of other files "
                  "from analyzer output:")
        LOG.debug(str(ex))

    LOG.debug("Collecting debug data")

    buildactions = [{
        'file': action.source,
        'command': action.original_command,
        'directory': action.directory}]

    for of in other_files:
        mentioned_file = os.path.abspath(os.path.join(action.directory, of))
        key = mentioned_file, action.target
        mentioned_file_action = actions_map.get(key)
        if mentioned_file_action is not None:
            buildactions.append({
                'file': mentioned_file_action.source,
                'command': mentioned_file_action.original_command,
                'directory': mentioned_file_action.directory})
        else:
            LOG.debug("Could not find %s in build actions.", key)

    from tu_collector import tu_collector

    tu_collector.zip_tu_files(zip_file, buildactions)

    # TODO: What about the dependencies of the other_files?
    tu_collector.add_sources_to_zip(
        zip_file,
        [os.path.join(action.directory, path) for path in other_files])

    with zipfile.ZipFile(zip_file, 'a') as archive:
        LOG.debug("[ZIP] Writing analyzer STDOUT to /stdout")
        archive.writestr("stdout", rh.analyzer_stdout)

        LOG.debug("[ZIP] Writing analyzer STDERR to /stderr")
        archive.writestr("stderr", rh.analyzer_stderr)

        LOG.debug("[ZIP] Writing extra information...")
        archive.writestr("build-action", action.original_command)
        archive.writestr(
            "analyzer-command",
            ' '.join([shlex.quote(x) for x in rh.analyzer_cmd]),
            )
        archive.writestr("return-code", str(rh.analyzer_returncode))

        toolchain = gcc_toolchain.toolchain_in_args(
            shlex.split(action.original_command))
        if toolchain:
            archive.writestr("gcc-toolchain-path", toolchain)

        compiler_info_file = os.path.join(rh.workspace, 'compiler_info.json')
        if os.path.exists(compiler_info_file):
            archive.write(compiler_info_file, "compiler_info.json")

    LOG.debug("ZIP file written at '%s'", zip_file)


def handle_failure(
    source_analyzer, rh, zip_file, result_base, actions_map, skip_handlers,
    rs_handler: ReviewStatusHandler
):
    """
    If the analysis fails a debug zip is packed together which contains
    build, analysis information and source files to be able to
    reproduce the failed analysis.
    """
    handle_reproducer(source_analyzer, rh, zip_file, actions_map)

    # In case of compiler errors the error message still needs to be collected
    # from the standard output by this postprocess phase so we can present them
    # as CodeChecker reports.
    checks = source_analyzer.config_handler.checks()
    state = checks.get('clang-diagnostic-error', (CheckerState.enabled, ''))[0]
    if state == CheckerState.enabled:
        rh.postprocess_result(skip_handlers, rs_handler)

    # Remove files that successfully analyzed earlier on.
    plist_file = result_base + ".plist"
    if os.path.exists(plist_file):
        os.remove(plist_file)


def kill_process_tree(parent_pid, recursive=False):
    """Stop the process tree try it gracefully first.

    Try to stop the parent and child processes gracefuly
    first if they do not stop in time send a kill signal
    to every member of the process tree.

    There is a similar function in the web part please
    consider to update that in case of changing this.
    """
    proc = psutil.Process(parent_pid)
    children = proc.children(recursive)

    # Send a SIGTERM (Ctrl-C) to the main process
    proc.terminate()

    # If children processes don't stop gracefully in time,
    # slaughter them by force.
    _, still_alive = psutil.wait_procs(children, timeout=5)
    for p in still_alive:
        p.kill()

    # Wait until this process is running.
    n = 0
    timeout = 10
    while proc.is_running():
        if n > timeout:
            LOG.warning("Waiting for process %s to stop has been timed out"
                        "(timeout = %s)! Process is still running!",
                        parent_pid, timeout)
            break

        time.sleep(1)
        n += 1


def setup_process_timeout(proc, timeout,
                          failure_callback=None):
    """
    Setup a timeout on a process. After `timeout` seconds, the process is
    killed by `signal_at_timeout` signal.

    :param proc: The subprocess.Process object representing the process to
      attach the watcher to.
    :param timeout: The timeout the process is allowed to run for, in seconds.
      This timeout is counted down some moments after calling
      setup_process_timeout(), and due to OS scheduling, might not be exact.
    :param failure_callback: An optional function which is called when the
      process is killed by the timeout. This is called with no arguments
      passed.

    :return: A function is returned which should be called when the client code
      (usually via subprocess.Process.wait() or
      subprocess.Process.communicate())) figures out that the called process
      has terminated. (See below for what this called function returns.)
    """

    # The watch dict encapsulated the captured variables for the timeout watch.
    watch = {'pid': proc.pid,
             'timer': None,  # Will be created later.
             'counting': False,
             'killed': False}

    def __kill():
        """
        Helper function to execute the killing of a hung process.
        """
        LOG.debug("Process %s has ran for too long, killing it!",
                  watch['pid'])
        watch['counting'] = False
        watch['killed'] = True
        kill_process_tree(watch['pid'], True)

        if failure_callback:
            failure_callback()

    timer = Timer(timeout, __kill)
    watch['timer'] = timer

    LOG.debug("Setup timeout of %s for PID %s", proc.pid, timeout)
    timer.start()
    watch['counting'] = True

    def __cleanup_timeout():
        """
        Stops the timer associated with the timeout watch if the process
        finished within the grace period.

        Due to race conditions and the possibility of the OS, or another
        process also using signals to kill the watched process in particular,
        it is possible that checking subprocess.Process.returncode on the
        watched process is not enough to see if the timeout watched killed it,
        or something else.

        (Note: returncode is -N where N is the signal's value, but only on Unix
        systems!)

        It is safe to call this function multiple times to check for the result
        of the watching.

        :return: Whether or not the process was killed by the watcher. If this
          is False, the process could have finished gracefully, or could have
          been destroyed by other means.
        """
        if watch['counting'] and not watch['killed'] and watch['timer']:
            watch['timer'].cancel()
            watch['timer'] = None
            watch['counting'] = False

        return watch['killed']

    return __cleanup_timeout


def collect_ctu_involved_files(result_handler, source_analyzer, output_dir):
    """
    This function collects the list of source files involved by CTU analysis.
    The list of files are written to output_dir.
    """
    if source_analyzer.ANALYZER_NAME != ClangSA.ANALYZER_NAME:
        return

    involved_files = set()

    involved_files.update(source_analyzer.get_analyzer_mentioned_files(
        result_handler.analyzer_stdout))
    involved_files.update(source_analyzer.get_analyzer_mentioned_files(
        result_handler.analyzer_stderr))

    out = os.path.join(output_dir, result_handler.analyzer_action_str)

    if involved_files:
        with open(out, 'w', encoding='utf-8', errors='ignore') as f:
            f.write('\n'.join(involved_files))
    elif os.path.exists(out):
        os.remove(out)


def check(check_data):
    """
    Invoke clang with an action which called by processes.
    Different analyzer object belongs to for each build action.

    skiplist handler is None if no skip file was configured.
    """
    actions_map, action, analyzer_config, \
        output_dir, skip_handlers, rs_handler, quiet_output_on_stdout, \
        capture_analysis_output, generate_reproducer, analysis_timeout, \
        ctu_reanalyze_on_failure, \
        output_dirs, statistics_data = check_data

    failed_dir = output_dirs["failed"]
    success_dir = output_dirs["success"]
    reproducer_dir = output_dirs["reproducer"]

    try:
        # If one analysis fails the check fails.
        return_codes = 0
        reanalyzed = False

        result_file = ''

        if analyzer_config is None:
            raise Exception("Analyzer configuration is missing.")

        source_analyzer, rh = prepare_check(action, analyzer_config,
                                            output_dir,
                                            skip_handlers, statistics_data)

        reanalyzed = os.path.exists(rh.analyzer_result_file)

        # Construct the analyzer cmd.
        analyzer_cmd = source_analyzer.construct_analyzer_cmd(rh)

        # The analyzer invocation calls __create_timeout as a callback
        # when the analyzer starts. This callback creates the timeout
        # watcher over the analyzer process, which in turn returns a
        # function, that can later be used to check if the analyzer quit
        # because we killed it due to a timeout.
        #
        # We need to capture the "function pointer" returned by
        # setup_process_timeout as reference, so that we may call it
        # later. To work around scoping issues, we use a list here so the
        # "function pointer" is captured by reference.
        timeout_cleanup = [lambda: False]

        if analysis_timeout and analysis_timeout > 0:
            def __create_timeout(analyzer_process):
                """
                Once the analyzer process is started, this method is
                called. Set up a timeout for the analysis.
                """
                timeout_cleanup[0] = setup_process_timeout(
                    analyzer_process, analysis_timeout)
        else:
            def __create_timeout(analyzer_process):
                # If no timeout is given by the client, this callback
                # shouldn't do anything.
                pass

        result_file_exists = os.path.exists(rh.analyzer_result_file)

        # Fills up the result handler with the analyzer information.
        source_analyzer.analyze(analyzer_cmd, rh, __create_timeout)

        # If execution reaches this line, the analyzer process has quit.
        if timeout_cleanup[0]():
            LOG.warning("Analyzer ran too long, exceeding time limit "
                        "of %d seconds.", analysis_timeout)
            LOG.warning("Considering this analysis as failed...")
            rh.analyzer_returncode = -1
            rh.analyzer_stderr = (">>> CodeChecker: Analysis timed out "
                                  "after {0} seconds. <<<\n{1}") \
                .format(analysis_timeout, rh.analyzer_stderr)

        source_analyzer.post_analyze(rh)

        # If source file contains escaped spaces ("\ " tokens), then
        # clangSA writes the plist file with removing this escape
        # sequence, whereas clang-tidy does not. We rewrite the file
        # names to contain no escape sequences for every analyzer.
        result_file = rh.analyzer_result_file.replace(r'\ ', ' ')
        result_base = os.path.basename(result_file)

        ctu_active = is_ctu_active(source_analyzer)

        zip_suffix = '_CTU' if ctu_active else ''

        failure_type = "_unknown"
        if rh.analyzer_returncode == 1:
            failure_type = "_compile_error"
        elif rh.analyzer_returncode == 254:
            failure_type = "_crash"
        elif rh.analyzer_returncode == 0:
            failure_type = ""

        zip_file = result_base + zip_suffix + failure_type + '.zip'
        failed_zip_file = os.path.join(failed_dir, zip_file)
        reproducer_zip_file = os.path.join(reproducer_dir, zip_file)

        return_codes = rh.analyzer_returncode

        source_file_name = os.path.basename(action.source)

        # Remove the previously generated .zip files.
        if os.path.exists(failed_zip_file):
            os.remove(failed_zip_file)
        if os.path.exists(reproducer_zip_file):
            os.remove(reproducer_zip_file)

        def handle_analysis_result(success, zip_file=zip_file):
            """
            Extra actions after executing the analysis. In case of success the
            clang-tidy output is transformed to reports. In case of failure the
            necessary files are collected for debugging. These .zip files can
            also be generated by --generate-reproducer flag.
            """
            if generate_reproducer:
                handle_reproducer(source_analyzer, rh,
                                  os.path.join(reproducer_dir, zip_file),
                                  actions_map)

            if success:
                handle_success(rh, result_file, result_base,
                               skip_handlers, rs_handler,
                               capture_analysis_output, success_dir)
            elif not generate_reproducer:
                handle_failure(source_analyzer, rh,
                               os.path.join(failed_dir, zip_file),
                               result_base, actions_map, skip_handlers,
                               rs_handler)

        if rh.analyzer_returncode == 0:
            handle_analysis_result(success=True)
            LOG.info("[%d/%d] %s analyzed %s successfully.",
                     progress_checked_num.value, progress_actions.value,
                     action.analyzer_type, source_file_name)

            if result_file_exists:
                LOG.debug("Previous analysis results in '%s' has been "
                          "overwritten.", rh.analyzer_result_file)
        else:
            LOG.error("Analyzing %s with %s %s failed!",
                      source_file_name,
                      action.analyzer_type,
                      "CTU" if ctu_active else "")

            if not quiet_output_on_stdout:
                LOG.error("\n%s", rh.analyzer_stdout)
                LOG.error("\n%s", rh.analyzer_stderr)

            handle_analysis_result(success=False)

            if ctu_active and ctu_reanalyze_on_failure:
                LOG.error("Try to reanalyze without CTU")
                # Try to reanalyze with CTU disabled.
                source_analyzer, rh = \
                    prepare_check(action, analyzer_config,
                                  output_dir,
                                  skip_handlers, statistics_data,
                                  True)
                reanalyzed = os.path.exists(rh.analyzer_result_file)

                # Construct the analyzer cmd.
                analyzer_cmd = source_analyzer.construct_analyzer_cmd(rh)

                # Fills up the result handler with
                # the analyzer information.
                source_analyzer.analyze(analyzer_cmd, rh)

                return_codes = rh.analyzer_returncode
                if rh.analyzer_returncode == 0:
                    handle_analysis_result(success=True)

                    LOG.info("[%d/%d] %s analyzed %s without"
                             " CTU successfully.",
                             progress_checked_num.value,
                             progress_actions.value,
                             action.analyzer_type,
                             source_file_name)

                    if result_file_exists:
                        LOG.debug("Previous analysis results in '%s' has "
                                  "been overwritten.",
                                  rh.analyzer_result_file)

                else:
                    LOG.error("Analyzing '%s' with %s without CTU failed.",
                              source_file_name, action.analyzer_type)

                    handle_analysis_result(success=False,
                                           zip_file=result_base + '.zip')

        collect_ctu_involved_files(rh, source_analyzer,
                                   output_dirs['ctu_connections'])

        if not quiet_output_on_stdout:
            if rh.analyzer_returncode:
                LOG.error('\n%s', rh.analyzer_stdout)
                LOG.error('\n%s', rh.analyzer_stderr)
            else:
                LOG.debug_analyzer('\n%s', rh.analyzer_stdout)
                LOG.debug_analyzer('\n%s', rh.analyzer_stderr)

        progress_checked_num.value += 1

        return return_codes, False, reanalyzed, action.analyzer_type, \
            result_file, action.source

    except Exception as e:
        LOG.debug_analyzer(str(e))
        traceback.print_exc(file=sys.stdout)
        return 1, False, reanalyzed, action.analyzer_type, None, \
            action.source


def skip_cpp(compile_actions, skip_handlers):
    """If there is no skiplist handler there was no skip list file in
       the command line.
       C++ file skipping is handled here.
    """

    if not skip_handlers:
        return compile_actions, []

    analyze = []
    skip = []
    for compile_action in compile_actions:

        if skip_handlers and skip_handlers.should_skip(compile_action.source):
            skip.append(compile_action)
        else:
            analyze.append(compile_action)

    return analyze, skip


def start_workers(actions_map, actions, analyzer_config_map,
                  jobs, output_path, skip_handlers,
                  rs_handler: ReviewStatusHandler, metadata_tool,
                  quiet_analyze, capture_analysis_output, generate_reproducer,
                  timeout, ctu_reanalyze_on_failure, statistics_data, manager,
                  compile_cmd_count):
    """
    Start the workers in the process pool.
    For every build action there is worker which makes the analysis.
    """
    # pylint: disable=no-member multiprocess module members.

    # Handle SIGINT to stop this script running.
    def signal_handler(signum, frame):
        try:
            pool.terminate()
            pool.join()
            manager.shutdown()
        except Exception as e:
            LOG.error("Failed to clean up after the pool!:\n")
            LOG.error(e)
            raise
        finally:
            sys.exit(128 + signum)

    actions, skipped_actions = skip_cpp(actions, skip_handlers)
    # Start checking parallel.
    checked_var = multiprocess.Value('i', 1)
    actions_num = multiprocess.Value('i', len(actions))
    pool = multiprocess.Pool(jobs,
                             initializer=init_worker,
                             initargs=(checked_var, actions_num))
    signal.signal(signal.SIGINT, signal_handler)

    # If the analysis has failed, we help debugging.
    failed_dir = os.path.join(output_path, "failed")
    if not os.path.exists(failed_dir):
        os.makedirs(failed_dir)

    # Analysis was successful processing results.
    success_dir = os.path.join(output_path, "success")
    if not os.path.exists(success_dir):
        os.makedirs(success_dir)

    # Similar to failed dir, but generated both in case of success and failure.
    reproducer_dir = os.path.join(output_path, "reproducer")
    if not os.path.exists(reproducer_dir) and generate_reproducer:
        os.makedirs(reproducer_dir)

    # Cppcheck raw output directory.
    cppcheck_dir = os.path.join(output_path, "cppcheck")
    if not os.path.exists(cppcheck_dir):
        os.makedirs(cppcheck_dir)

    # Collect what other TUs were involved during CTU analysis.
    ctu_connections_dir = os.path.join(output_path, "ctu_connections")
    if not os.path.exists(ctu_connections_dir):
        os.makedirs(ctu_connections_dir)

    output_dirs = {'success': success_dir,
                   'failed': failed_dir,
                   'reproducer': reproducer_dir,
                   'ctu_connections': ctu_connections_dir}

    analyzed_actions = [(actions_map,
                         build_action,
                         analyzer_config_map.get(build_action.analyzer_type),
                         output_path,
                         skip_handlers,
                         rs_handler,
                         quiet_analyze,
                         capture_analysis_output,
                         generate_reproducer,
                         timeout,
                         ctu_reanalyze_on_failure,
                         output_dirs,
                         statistics_data)
                        for build_action in actions]

    if analyzed_actions:
        try:

            # Workaround, equivalent of map.
            # The main script does not get signal
            # while map or map_async function is running.
            # It is a python bug, this does not happen if a timeout is
            # specified, then receive the interrupt immediately.

            # FIXME: Ensure all shared data structures are wrapped in manager
            #        proxy objects before passing them to other processes via
            #        map_async.
            #        Note that even deep-copying is known to be insufficient.
            timeout = 3155760 if sys.platform == 'win32' else 31557600
            pool.map_async(check,
                           analyzed_actions,
                           1,
                           callback=lambda results: worker_result_handler(
                               results, metadata_tool, output_path)
                           ).get(timeout)

            pool.close()
        except Exception:
            pool.terminate()
            raise
        finally:
            pool.join()
    else:
        pool.close()
        pool.join()
        LOG.info("----==== Summary ====----")

    for skp in skipped_actions:
        LOG.debug_analyzer("%s is skipped", skp.source)

    LOG.info("Total analyzed compilation commands: %d",
             compile_cmd_count.analyze)
    # Some compile commands are skipped during log processing, if nothing
    # is skipped there for pre-analysis step, files will be skipped only
    # during analysis.
    if skipped_actions or compile_cmd_count.skipped:
        LOG.info("Skipped compilation commands: %d",
                 compile_cmd_count.skipped + len(skipped_actions))

    LOG.info("----=================----")
    if not os.listdir(success_dir):
        shutil.rmtree(success_dir)

    if not os.listdir(failed_dir):
        shutil.rmtree(failed_dir)



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import re


def available(ordered_checkers, available_checkers):
    """Verify if every element in the ordered checkers a valid checker name.

    Every element in the ordered checkers should match partially
    a checker or profile name in available_checkers.

    Returns the set of checker names without any match.
    """
    missing_checkers = set()
    for checker_name, _ in ordered_checkers:
        # TODO: This label list shouldn't be hard-coded here.
        if checker_name.startswith('profile:') or \
                checker_name.startswith('guideline:') or \
                checker_name.startswith('severity:') or \
                checker_name.startswith('sei-cert:'):
            continue

        name_match = False
        for available_checker in available_checkers:
            regex = "^" + re.escape(str(checker_name)) + ".*$"
            c_name = available_checker
            match = re.match(regex, c_name)
            if match:
                name_match = True
                break

        if not name_match:
            missing_checkers.add(checker_name)
    return missing_checkers



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

""" Convert between Report type and thrift ReportData type. """

from typing import Callable, List

from codechecker_api.codeCheckerDBAccess_v6.ttypes import \
    ExtendedReportDataType, ReportData, Severity

from codechecker_report_converter.report import BugPathEvent, \
    BugPathPosition, File, MacroExpansion, Range, Report


def to_report(
    report: ReportData,
    get_file: Callable[[int, str], File]
) -> Report:
    """ Create a Report object from the given thrift report data. """
    severity = Severity._VALUES_TO_NAMES[report.severity] \
        if report.severity else 'UNSPECIFIED'

    bug_path_events: List[BugPathEvent] = []
    bug_path_positions: List[BugPathPosition] = []
    notes: List[BugPathEvent] = []
    macro_expansions: List[MacroExpansion] = []

    details = report.details
    if details:
        for e in details.pathEvents:
            bug_path_events.append(BugPathEvent(
                e.msg,
                get_file(e.fileId, e.filePath),
                e.startLine,
                e.startCol,
                Range(e.startLine, e.startCol, e.endLine, e.endCol)))

        for p in details.executionPath:
            bug_path_positions.append(BugPathPosition(
                get_file(p.fileId, p.filePath),
                Range(p.startLine, p.startCol, p.endLine, p.endCol)))

        for e in details.extendedData:
            if e.type == ExtendedReportDataType.NOTE:
                notes.append(BugPathEvent(
                    e.message,
                    get_file(e.fileId, e.filePath),
                    e.startLine,
                    e.startCol,
                    Range(e.startLine, e.startCol, e.endLine, e.endCol)))

            if e.type == ExtendedReportDataType.MACRO:
                name = ''
                macro_expansions.append(MacroExpansion(
                    e.message,
                    name,
                    get_file(e.fileId, e.filePath),
                    e.startLine,
                    e.startCol,
                    Range(e.startLine, e.startCol, e.endLine, e.endCol)))

    return Report(
        get_file(report.fileId, report.checkedFile),
        report.line,
        report.column,
        report.checkerMsg,
        report.checkerId,
        severity,
        report.bugHash,
        report.analyzerName,
        bug_path_events=bug_path_events or None,
        bug_path_positions=bug_path_positions,
        notes=notes,
        macro_expansions=macro_expansions)


def to_report_data(
    report: Report
) -> ReportData:
    """ Convert a Report object to a Thrift ReportData type. """
    severity = Severity._NAMES_TO_VALUES[report.severity or 'UNSPECIFIED']
    return ReportData(
        checkerId=report.checker_name,
        bugHash=report.report_hash,
        checkedFile=report.file.path,
        checkerMsg=report.message,
        line=report.line,
        column=report.column,
        severity=severity,
        analyzerName=report.analyzer_name,
        bugPathLength=len(report.bug_path_events))



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------
"""
Product module.
"""


from codechecker_common.logger import get_logger

LOG = get_logger('system')


def __strip_protocol_from_url(url):
    """
    Strip the PROTOCOL specifier from an URL string and return it, along with
    the remainder of the URL.
    """
    if '://' not in url:
        return None, url

    parts = url.split('://', 1)
    protocol = parts[0]
    url = url.replace(parts[0] + '://', '').lstrip('/').rstrip('/')

    return protocol, url


def expand_whole_protocol_and_port(protocol=None, port=None):
    """
    Calculate a full protocol and port value from the possibly None protocol
    and port values given. This method helps to default port numbers to
    connection protocols understood by CodeChecker.
    """

    if protocol:
        if protocol == 'http':
            portnum = 80
        elif protocol == 'https':
            portnum = 443
        else:
            raise ValueError("'{0}' is not a protocol understood by "
                             "CodeChecker".format(protocol))
    else:
        protocol = 'http'
        portnum = 8001

    if port:
        portnum = port

    return protocol, portnum


def understand_server_addr(server_addr):
    """
    Attempts to understand the given server address and fetch the hostname
    or host address and optionally the port from it.
    """
    parts = server_addr.split(':')

    if len(parts) == 2:
        return parts[0], int(parts[1])
    if len(parts) == 1:
        return parts[0], None

    # Multiple ':'s found in the URL, which means it could be an
    # IPv6 address, if it is in the right format.
    if parts[0].startswith('['):
        if parts[-1].endswith(']'):
            # The address was [::1], there is no port number.
            return ':'.join(parts), None
        if parts[-2].endswith(']'):
            # The address was such as [::1]:1234.
            return ':'.join(parts[:-1]), int(parts[-1])

    raise ValueError("The server's address is not in a valid "
                     "'host:port' or '[host]:port' format!")


def split_server_url(url):
    """
    Splits the given CodeChecker server URL into its parts.

    The format of a valid URL is:
      protocol://host:port/
    where
      * Protocol: HTTP or HTTPS
      * Host: The server's host name or IP address
      * Port: The server's port number

    As a shortcut, the following formats are also valid:
      hostname           (means: http://hostname:8001)
    """

    LOG.debug("Parsing server url '%s'", url)
    protocol, url = __strip_protocol_from_url(url)

    # A valid product_url looks like this: 'http://localhost:8001/Product'.
    protocol, port = expand_whole_protocol_and_port(protocol, None)
    host = 'localhost'
    try:
        parts = url.split('/', 1)

        # Something is either a hostname, or a host:port.
        server_addr = parts[0]
        host, maybe_port = understand_server_addr(server_addr)
        if maybe_port:
            port = maybe_port
    except Exception:
        raise ValueError("The specified server URL is invalid.")

    LOG.debug("Result: With '%s' on server '%s:%s'",
              protocol, host, port)

    return protocol, host, port


def create_product_url(protocol, host, port, endpoint):
    return "{0}://{1}:{2}{3}".format(protocol, host, str(port), endpoint)


def split_product_url(url):
    """
    Splits the given CodeChecker server's product-specific URL into its parts.

    The format of a valid URL is:
      protocol://host:port/ProductEndpoint
    where
      * Protocol: HTTP or HTTPS
      * Host: The server's host name or IP address
      * Port: The server's port number
      * ProductEndpoint: The product's unique endpoint folder under the server.

    As a shortcut, the following formats are also valid:
      ProductEndpoint           (means: http://localhost:8001/ProductEndpoint)
      hostname/ProductEndpoint  (means: http://hostname:8001/ProductEndpoint)
    """

    LOG.debug("Parsing product url '%s'", url)
    protocol, url = __strip_protocol_from_url(url)

    # A valid product_url looks like this: 'http://localhost:8001/Product'.
    protocol, port = expand_whole_protocol_and_port(protocol, None)
    host, product_name = 'localhost', 'Default'
    try:
        parts = url.split("/")

        if len(parts) == 1:
            # If only one word is given in the URL, consider it as product
            # name, but then it must appear to be a valid product name.

            # "Only one word" URLs can be just simple host names too:
            # http://codechecker.example.com:1234 should NOT be understood as
            # the "codechecker.example.com:1234" product on "localhost:8001".
            product_name = parts[0]
            if product_name[0].isdigit() or '.' in product_name \
                    or ':' in product_name:
                raise ValueError("The given product URL is invalid. Please "
                                 "specify a full product URL.")
        elif len(parts) == 2:
            # URL is at least something/product-name.
            product_name = parts[1]

            # Something is either a hostname, or a host:port.
            server_addr = parts[0]
            host, maybe_port = understand_server_addr(server_addr)
            if maybe_port:
                port = maybe_port
        else:
            raise ValueError("Product URL can not contain extra '/' chars.")
    except Exception:
        raise ValueError("The specified product URL is invalid.")

    LOG.debug("Result: With '%s' on server '%s:%s', product '%s'",
              protocol, host, port, product_name)

    return protocol, host, port, product_name




# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

"""
Map between database statuses and the corresponding messages.
"""


from codechecker_api_shared.ttypes import DBStatus

db_status_msg = {
    DBStatus.OK: "Database is up to date.",
    DBStatus.MISSING: "Database is missing.",
    DBStatus.FAILED_TO_CONNECT: "Connection failed to the database server.",
    DBStatus.SCHEMA_MISMATCH_OK:
        "Database schema mismatch! Possible to upgrade.",
    DBStatus.SCHEMA_MISMATCH_NO:
        "Database schema mismatch! NOT possible to upgrade.",
    DBStatus.SCHEMA_MISSING: "Database schema is missing.",
    DBStatus.SCHEMA_INIT_ERROR: "Schema initialization error.",
    DBStatus.SCHEMA_UPGRADE_FAILED: "Schema upgrade failed."
}



# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

import json
import logging
import os

from typing import Dict, List, Optional

from codechecker_report_converter.report import BugPathEvent, File, \
    get_or_create_file, Report

from ..analyzer_result import AnalyzerResultBase


LOG = logging.getLogger('report-converter')


class AnalyzerResult(AnalyzerResultBase):
    """ Transform analyzer result of the FB Infer. """

    TOOL_NAME = 'fbinfer'
    NAME = 'Facebook Infer'
    URL = 'https://fbinfer.com'

    def __init__(self):
        super(AnalyzerResult, self).__init__()
        self.__infer_out_parent_dir = None
        self.__file_cache: Dict[str, File] = {}

    def get_reports(self, result_file_path: str) -> List[Report]:
        """ Parse the given analyzer result. """
        reports: List[Report] = []

        if os.path.isdir(result_file_path):
            report_file = os.path.join(result_file_path, "report.json")
            self.__infer_out_parent_dir = os.path.dirname(result_file_path)
        else:
            report_file = result_file_path
            self.__infer_out_parent_dir = os.path.dirname(
                os.path.dirname(result_file_path))

        if not os.path.exists(report_file):
            LOG.error("Report file does not exist: %s", report_file)
            return reports

        try:
            with open(report_file, 'r',
                      encoding="utf-8", errors="ignore") as f:
                bugs = json.load(f)
        except IOError:
            LOG.error("Failed to parse the given analyzer result '%s'. Please "
                      "give an infer output directory which contains a valid "
                      "'report.json' file.", result_file_path)
            return reports

        for bug in bugs:
            report = self.__parse_report(bug)
            if report:
                reports.append(report)

        return reports

    def __get_abs_path(self, source_path):
        """ Returns full path of the given source path.
        It will try to find the given source path relative to the given
        analyzer report directory (infer-out).
        """
        if os.path.exists(source_path):
            return os.path.abspath(source_path)

        full_path = os.path.join(self.__infer_out_parent_dir, source_path)
        if os.path.exists(full_path):
            return full_path

        LOG.warning("No source file found: %s", source_path)

    def __parse_report(self, bug) -> Optional[Report]:
        """ Parse the given report and create a message from them. """
        report_hash = bug['hash']
        checker_name = bug['bug_type']

        message = bug['qualifier']
        line = int(bug['line'])
        col = int(bug['column'])
        if col < 0:
            col = 0

        source_path = self.__get_abs_path(bug['file'])
        if not source_path:
            return None

        report = Report(
            get_or_create_file(
                os.path.abspath(source_path), self.__file_cache),
            line, col, message, checker_name,
            report_hash=report_hash,
            bug_path_events=[])

        for bug_trace in bug['bug_trace']:
            event = self.__parse_bug_trace(bug_trace)

            if event:
                report.bug_path_events.append(event)

        report.bug_path_events.append(BugPathEvent(
            report.message, report.file, report.line, report.column))

        return report

    def __parse_bug_trace(self, bug_trace) -> Optional[BugPathEvent]:
        """ Creates event from a bug trace element. """
        source_path = self.__get_abs_path(bug_trace['filename'])
        if not source_path:
            return None

        message = bug_trace['description']
        line = int(bug_trace['line_number'])
        col = int(bug_trace['column_number'])
        if col < 0:
            col = 0

        return BugPathEvent(
            message,
            get_or_create_file(source_path, self.__file_cache),
            line,
            col)
